(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

signature ABSYNUTILS =
sig

(* by CfaConstruct ***************************************************)
val varName: Absyn.varid -> string
val varLName: Absyn.varlongid -> string
val strLName: Absyn.strlongid -> string
val printTy: Ty.ty -> unit
val isFnty: Ty.ty -> bool
val hasFnty: Ty.ty -> bool
val isExnty: Ty.ty -> bool
val isExn: Ty.ty -> bool
val hasExnty: Ty.ty -> bool
val ioTy: Ty.ty -> bool * bool
val outTy: Ty.ty -> int -> bool * bool
val appTy: Ty.ty -> int -> Ty.ty
val clearTy: Ty.ty -> Ty.ty
val unifiable: Ty.ty -> Ty.ty -> bool
val getLoc: LocationInfo.info -> Location.t
val patName: Absyn.pat -> string
val init: unit -> unit

(* by EaConstruct ***************************************************)
val dummyInfo: LocationInfo.info
val conLName: Absyn.conlongid -> string
val loc2str: Location.t -> string

end (* ABSYNUTILS *)

structure AbsynUtils : ABSYNUTILS =
struct

open Absyn
open Utils
structure L = List
structure TU = TyUtil

(* by CfaConstruct ***************************************************)
(* varName: varid -> string *)
fun varName (name,_) = name

(* varLName: varlongid -> string *)
fun varLName (strids, (name,_), _) =
  (foldl (fn res (sid,_) => res^(sid^".")) "" strids)^name

(* strLName: strlongid -> string *)
val strLName = varLName

(* printTy: ty -> unit *)
val printTy = TU.print_ty

(* isFnty: ty -> bool *)
val isFnty = TU.is_fun

(* hasFnty: ty -> bool *)
val hasFnty = TU.has_fun

(* isExnty: ty -> bool *)
val isExnty = TU.is_exn_type

(* isExn: ty -> bool *)
val isExn = TU.is_exn

(* ty |-> flag *)
(* cache of hasExnty *)
val t2fT = ref TM.empty
(* hasExnty: ty -> bool *)
fun hasExnty ty =
  TM.find ty (!t2fT) handle Not_found =>
  let val _ = t2fT := TM.add ty false (!t2fT)
      val flg = case ty
                  of Ty.Var _ => true
                   | Ty.Fun (t1,t2) =>
                     (isExn t2) orelse (hasExnty t1) orelse (hasExnty t2)
                   | Ty.Rec (flds,_) => L.exists (fn (_,t) => hasExnty t) flds
                   | Ty.Data (tys, ("list",_,_,_)) =>
                     (isExn ty) orelse (L.exists hasExnty tys)
                   | Ty.Data (tys, tyname) =>
                     (isExn ty) orelse (L.exists hasExnty tys)
                     orelse (L.exists (fn (_,None) => TU.is_abs ty
                                        | (_,Some t) => hasExnty t)
                            (Ty.CE.find tyname))
                   | Ty.Tuple tys => L.exists hasExnty tys
                   | Ty.Link t => hasExnty (!t)
  in  (t2fT := TM.add ty flg (!t2fT); flg)
  end

(* ioTy: Ty.ty -> bool * bool *)
fun ioTy ty =
  case ty of Ty.Fun (t1, t2) => (hasExnty t1, hasExnty t2)
           | Ty.Link t => ioTy (!t)
           | Ty.Var _ => (true, true)
           | _ => if TU.is_abs ty then (true, true)
                  else (print "\n"; printTy ty; raise Fail "ioTy:no ftn type")

(* outTy: ty -> int -> bool * bool *)
fun outTy t cnt =
  let val oty = TU.app_type t cnt in (isFnty oty, isExnty oty) end
  handle _ => if TU.is_abs t then (true, true)
              else (print "\n"; printTy t; raise Fail "outTy:no ftn type")

(* appTy: ty -> int -> ty *)
val appTy = TU.app_type

(* clearTy: Ty.ty -> Ty.ty *)
fun clearTy ty = let val (_, t) = TU.clear_type ty in t end

(* unifiable: Ty.ty -> Ty.ty -> bool *)
val unifiable = TU.unifiable

(* getLoc: LocationInfo.info -> Location.t *)
fun getLoc loc = loc

(* patName: pat -> string *)
fun patName pat = case pat of VarPat (vid, _, _) => varName vid
                            | AsPat (vid, _, _, _) => varName vid
                            | _ => raise Fail "AbsynUtils: patName"

(* init: unit -> unit *)
fun init() = t2fT := TM.empty

(* by EaConstruct ***************************************************)
open Location

val dummyInfo = none

(* conLongName: conlongid -> string *)
fun conLName (strids, (name,_), _) =
  (foldl (fn res (sid,_) => res^(sid^".")) "" strids)^name

(* loc2str: Location.t -> string *)
fun loc2str loc =
  if String.length !input_name = 0
  then ((i2s loc.loc_start)^"-"^(i2s loc.loc_end))
  else let val (filename, linenum, linebeg) =
               Linenum.for_position !input_name loc.loc_start
       in  (filename^":"^(i2s linenum)^"."^(i2s(loc.loc_start - linebeg))
            ^("-"^(i2s(loc.loc_end - linebeg))))
       end

(*
fun loc2str loc =
  if String.length !input_name = 0
  then ((i2s loc.loc_start)^"-"^(i2s loc.loc_end))
  else let val (filename, linenum, linebeg) =
               Linenum.for_position !input_name loc.loc_start
       in  ("File \""^filename^"\", line "^(i2s linenum)^", characters "
            ^(i2s (loc.loc_start - linebeg))^"-"
            ^(i2s (loc.loc_end - linebeg)))
       end
*)

end (* AbsynUtils *)
