(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

signature CFACONSTRUCT =
sig
val construct: Absyn.topdec -> CfaUtils.basis
end (* CFACONSTRUCT *)

structure CfaConstruct : CFACONSTRUCT =
struct

open Absyn
open Utils
open AbsynUtils
open CfaUtils
structure L = List
structure TU = TyUtil

(* lid list -> lid list *)
fun filtDum lids = L.filter (fn i => i > 0) lids

(* nm2ty: varidinfo -> vel -> Ty.ty *)
fun nm2ty nm nil = raise Fail "strange ValueEnvironment"
  | nm2ty nm ((nm',(_,ty))::rest) = if nm=nm' then ty else nm2ty nm rest

(* mkWild: int -> string -> string *)
fun mkWild cnt res = if cnt = 0 then res else mkWild (cnt-1) (" _"^res)

(* dec: env -> dec -> env *)
fun dec env d =
  case d
    of ValDec (_, vbs, _, _) => (* vbs: val and *)
       let val oft = env.ftns val oow = env.owns
	   val (nft, now) =
	       foldl (fn (ft,ow) (p,e,_) =>
		         let val ev = vb env p e (expTy e) nil
			 in  (fPushs (fPops ev.ftns oft) ft,
			      oPushs (oPops ev.owns oow) ow)
			 end) (ftns0,owns0) vbs
       in  (env{ftns <- fPushs nft oft}){owns <- oPushs now oow}
       end
     | RecValDec (_, vbs, vel, _) =>
       let fun mkFs nil res = res
             | mkFs ((p,e as FnExp((ps,_,_)::_,_,_),_)::rest) (vbs',ftn') =
	       let val fid = nFid() val _ = mkFids ((L.length ps)-1)
		   val fnm = patName p val lid = nLid()
	       in  mkFs rest (vbs'@[(e,nm2ty fnm vel,fid,lid,fnm)],
			      fPush (fid,lid,fnm) ftn')
	       end
             | mkFs _ _ = raise Fail "strange RecValDec"
           val (nvbs, addf) = mkFs vbs (nil, ftns0)
	   val nev = env{ftns <- fPushs addf env.ftns}
       in  (L.iter (fn (FnExp(rules as (ps,e,_)::_,_,i),t,fid,lid,fname) =>
		       let val numPat = L.length ps val loc = getLoc i
			   val bodies = fnexp nev rules fid t [fname] loc
		       in  (addFids (fid,lid,fname,loc,numPat,ioTy t);
			    if numPat > 1 orelse isFnty(expTy e)
			    then addLa fid (filtDum bodies);
			    addF lid fid; addT t fid)
		       end
                     | _ => raise Fail "strange RecValDec") nvbs;
            nev)
       end
     | LocalDec (d1, d2, _) => (* get rid of local dec. env. *)
       let val ev1 = dec env d1 val ev2 = dec ev1 d2
	   val addf = fPops ev2.ftns ev1.ftns
	   val addo = oPops ev2.owns ev1.owns
       in  (env{ftns <- fPushs addf env.ftns}){owns <- oPushs addo env.owns}
       end
     | SeqDec (ds, _) => foldl dec env ds
     | _ => env

(* vb: env -> pat -> exp -> ty -> string list -> env *)
and vb env p e t names =
  if isFnty t
  then case p
         of WildPat _ => (exp env e; env)
          | VarPat (vid,_,_) => vbexp env ((varName vid)::names) e t
          | AsPat (vid,p',_,_) => vb env p' e t ((varName vid)::names)
          | _ => raise Fail "type checking"
  else (exp env e; env{owns <- oPushs (mkOwn (-1) (-1) p) env.owns})

(* vbexp: env -> string list -> exp -> ty -> env *)
and vbexp env fnames e t =
  case e
    of FnExp (rules as (ps,e',_)::_, _, info) =>
       let val argnum = L.length ps val loc = getLoc info
           val fid = nFid() val _ = mkFids (argnum-1)
           val bodies = fnexp env rules fid t fnames loc
           val lid = nLid() val addf = L.map (fn nm => (fid,lid,nm)) fnames
       in  (addFids' (fid,lid,fnames,loc,argnum,ioTy t);
            if argnum > 1 orelse isFnty(expTy e')
            then addLa fid (filtDum bodies); addF lid fid; addT t fid;
            env{ftns <- fPushs addf env.ftns})
           end
     | LetExp (d, e', _, info) => (* get rid of let dec. env. *)
       let val ev1 = dec env d val ev2 = vbexp ev1 fnames e' t
           val addf = fPops ev2.ftns ev1.ftns
           val addo = oPops ev2.owns ev1.owns
       in  (env{ftns <- fPushs addf env.ftns}){owns <- oPushs addo env.owns}
       end
     | VarExp (vlid, _, info) =>
       let val name = varLName vlid val fid = nFid()
       in  case fnm2fid vlid env.ftns
             of Some (f,_) =>
                let val lid = nLid() val loc = getLoc info
                    val addf = L.map (fn nm => (fid,lid,nm)) fnames
                in  (addFids' (fid,lid,fnames,loc,getAri f,ioTy t);
                     addLa fid (getLa f); addF lid fid; addF lid f;
                     addLoc2lid loc lid;
                     addT t fid; env{ftns <- fPushs addf env.ftns})
                end
              | None =>
                let val lid = exp env e val loc = getLoc info
                    val addf = L.map (fn nm => (fid,lid,nm)) fnames
                    val _ = addLoc2lid loc lid
                in  if lid = -1
                    then if isExnty t then env else raise Fail "getLid"
                    else (addFids' (fid,lid,fnames,loc,-1,ioTy t);
                          if (isFnty(appTy t 1) handle _ =>
                              if TU.is_abs t then true else false)
                          then addLb lid fid;
                          addF lid fid; addT t fid;
			  env{ftns <- fPushs addf env.ftns})
                end
       end
     | AppExp (_, _, _, info) =>
       let val fid = nFid() val lid = exp env e val loc = getLoc info
           val (lid1, flg) =
               case getLid2le lid of Some (Ae l1) => (l1, false)
                                   | Some (FAe l1) => (l1, true)
                                   | _ => raise Fail "AppExp"
           val (fid', org) = getHifn lid1 env.ftns
           val argnum = (getAri org) handle Fail "getAri" => -1
           val addf = L.map (fn nm => (fid,lid,nm)) fnames
       in  if lid = -1
           then if isExnty t then env else raise Fail "getLid"
           else (addFids' (fid,lid,fnames,loc,argnum,ioTy t);
                 if flg then addLa fid (getLa (fid'+1))
                 else if (isFnty (appTy t 1) handle _ =>
                          if TU.is_abs t then true else false)
                 then addLb lid fid;
                 addLoc2lid loc lid;
                 addF lid fid; addT t fid; env{ftns <- fPushs addf env.ftns})
       end
     | _ => let val fid = nFid() val lid = exp env e val loc = expLoc e
                val addf = L.map (fn nm => (fid,lid,nm)) fnames
            in  if lid = -1
                then if isExnty t then env else raise Fail "getLid"
                else (addFids' (fid,lid,fnames,loc,-1,ioTy t);
                      if (isFnty (appTy t 1) handle _ =>
                          if TU.is_abs t then true else false)
                      then addLb lid fid;
                      addLoc2lid loc lid; addF lid fid; addT t fid;
		      env{ftns <- fPushs addf env.ftns})
            end

(* fnexp: environ -> fnrule list -> fid -> ty -> string list *)
(*      -> Location.t -> lid list *)
(* env: ftns are set but owner is not yet *)
(* ouput: lid list of bodies *)
and fnexp env rules fid ty fnames loc =
  case rules
    of nil => raise Fail "empty frules"
     | ([p],_,_)::rest => (* one argument *)
       L.map (fn ([pat],e,_) =>
                 exp (env{owns <- oPushs (mkOwn fid 0 pat) env.owns}) e
               | _ => raise Fail "strange frules") rules
     | (pats,_,_)::rest =>  (* multiple arguments *)
       let val numPat = L.length pats
           val (isFnOty, isExnOty) = outTy ty numPat
           val fids = mkFids' (numPat-1) fid nil
           fun rule (ps,e,_) =
             let val addo = L.fold_left2 (fn r f p => oPushs r (mkOwn f 0 p))
                                         owns0 (fid::fids) ps
             in  exp (env{owns <- oPushs addo env.owns}) e
             end
           val bodies = L.map rule rules
           val body = if isFnOty then addE (filtDum bodies) isExnOty else -1
           val lids = mkLids (numPat-1) nil
           fun mkFnm f =
             L.map (fn nm => ("/"^nm^(mkWild (f-fid) "")^" /")) fnames
           val infos = L.map2 (fn l f =>
                               let val nms = mkFnm f val t = appTy ty (f-fid)
                               in  (addFids'(f,l,nms,loc,fid-f+numPat,ioTy t);
                                    (f,t,l))
                               end) lids fids
       in  [foldl (fn l (fid',fty,lid) =>
                      (if l > 0 then addLa fid' [l]; addF lid fid';
                       addT fty fid'; lid)) body (L.rev infos)]
       end

(* exp: env -> exp -> lid *)
and exp env e =
  case e
    of VarExp (vlid, t, info) =>
       if isFnty t
       then let val name = varLName vlid
            in  case fnm2fid vlid env.ftns
                  of Some (fid,lid) => lid
                   | None =>
                     let val lid = nLid()
                         val (own,pos) = getOwner name env.owns
                     in (if own = -1 orelse pos < 0
                         then lid2le lid (De t)
                         else (addAb lid own pos t;
                               lid2le lid (Ve (vlid,own,pos))); lid)
                     end
            end
       else -1
     | AppExp (ConExp _, e', _, _) => (ignore(exp env e'); -1)
     | AppExp (e1, e2, t, _) =>
       let val lid1 = exp env e1
           val lids = case e2
                        of TupleExp (es,_,_) => L.map (exp env) es
                         | _ => [exp env e2]
           val lid = nLid() val loc1 = expLoc e1 val t' = expTy e2
       in (if hasFnty t' then addApAb lid1 (filtDum lids);
           result := (loc1, lid1)::(!result);
           addLoc2lid loc1 lid1;
           if isFnty t
           then (addApL lid lid1;
                 let val (fid, org) = getHifn lid1 env.ftns
                     val argnum = (getAri org) handle Fail "getAri" => -1
                 in  if argnum <> -1
                     then if fid > 0 andalso (fid+1) < (org+argnum)
                          then (addF lid (fid+1); lid2le lid (FAe lid1); lid)
                          else if fid < 0 andalso (fid-1) > (org-argnum)
                               then (addF lid (fid-1);
				     lid2le lid (FAe lid1);lid)
                               else (lid2le lid (Ae lid1);lid)
                     else (lid2le lid (Ae lid1);lid)
                 end)
           else (addDmAp lid; -1))
       end
     | RecordExp (labexps, _, _) =>
       (L.iter (fn (_,e',_) => ignore(exp env e')) labexps; -1)
     | RecordFieldExp (e', _, t, _) =>
       (ignore(exp env e');
        if isFnty t
        then let val lid = nLid() in (lid2le lid (De t); lid) end
        else -1)
     | ArrayFieldExp (e1, e2, t, _) =>
       (ignore(exp env e1); ignore(exp env e2);
        if isFnty t
        then let val lid = nLid() in (lid2le lid (De t); lid) end
        else -1)
     | UpdateArrayExp (e1, e2, e3, _, _) =>
       (L.iter (ignore $$ (exp env)) [e1, e2, e3]; -1)
     | SubstRecordExp (e1, _, e2, _, _) =>
       (L.iter (ignore $$ (exp env)) [e1, e2]; -1)
     | TupleExp (es, _, _) => (L.iter (ignore $$ (exp env)) es; -1)
     | ListExp (es, _, _) => (L.iter (ignore $$ (exp env)) es; -1)
     | ArrayExp (es, _, _) => (L.iter (ignore $$ (exp env)) es; -1)
     | LetExp (d, e', t, _) =>
       let val ev = dec env d val lid' = exp ev e'
       in  if isFnty t then addE (filtDum [lid']) (isExnty t) else -1
       end
     | HandleExp (e', rules, t, info) =>
       if isFnty t
       then let val fid1 = nFid() (* handlee *)
                val fid2 = nFid() (* handler *)
                val lid1 = exp env e'
                val lids = L.map (fn (_,e'',_) => exp env e'') rules
                val lid2 = nLid() val loc = getLoc info
            in  (addFids (fid2,lid2,"/handler ftn/",loc,1,(true,hasExnty t));
                 addF lid2 fid2;  (* handlers not collected by types *)
                 addE (filtDum (lid1::lids)) (isExnty t)) end
       else let val fid2 = (ignore(nFid()); nFid())
                val _ = (ignore(exp env e');
                         L.iter (fn (_,e'',_) => ignore(exp env e'')) rules)
                val lid2 = nLid() val loc = getLoc info
            in  (addFids (fid2,lid2,"/handler ftn/",loc,1,(true,hasExnty t));
                 addF lid2 fid2; -1)
            end
     | RaiseExp (e', _, _) => (ignore(exp env e'); -1)
     | FnExp (rules as (ps,e',_)::_, t, info) =>
       let val argnum = L.length ps val loc = getLoc info
           val fid = nFid() val _ = mkFids (argnum-1)
           val bodies = fnexp env rules fid t ["/unnamed ftn/"] loc
           val lid = nLid()
       in  (addFids (fid,lid,"/unnamed ftn/",loc,argnum,ioTy t);
            if argnum > 1 orelse isFnty(expTy e')
            then addLa fid (filtDum bodies); addF lid fid; addT t fid; lid)
       end
     | AssignExp (e1, e2, _, _) => (L.iter (ignore $$ (exp env)) [e1, e2]; -1)
     | RefExp (e', _, _) => (ignore(exp env e'); -1)
     | DeRefExp (e', t, _) =>
       (ignore(exp env e');
        if isFnty t
        then let val lid = nLid() in (lid2le lid (De t); lid) end
        else -1)
     | SeqExp (es, t, _) =>
       if isFnty t
       then let val lidn = foldl (fn l e => exp env e) 0 es
            in  addE (filtDum [lidn]) (isExnty t)
            end
       else (L.iter (ignore $$ (exp env)) es; -1)
     | CaseExp (e', rules, t, _) =>
       let fun mko p = env{owns <- oPushs (mkOwn (-1) (-1) p) env.owns}
           val _ = exp env e'
           val lids = L.map (fn (p,e'',_) => exp (mko p) e'') rules
       in  if isFnty t then addE (filtDum lids) (isExnty t) else -1
       end
     | IfExp (e1, e2, e3, t, _) =>
       if isFnty t
       then let val _ = exp env e1 val lid2 = exp env e2 val lid3 = exp env e3
            in  addE (filtDum [lid2,lid3]) (isExnty t)
            end
       else (L.iter (ignore $$ (exp env)) [e1, e2, e3]; -1)
     | WhileExp (e1, e2, _, _) => (L.iter (ignore $$ (exp env)) [e1, e2]; -1)
     | ForExp (_, e1, e2, e3, e4, _, _) =>
       (L.iter (ignore $$ (exp env)) [e1, e2, e3, e4]; -1)
     | _ => -1

(* strexp: fc -> gc -> ctxt -> strexp -> string -> ctxt *)
fun strexp fc gc (c as C(sc,ci)) stre name =
  case stre
    of VarStr (strlid as (strids, strid, _), _) =>
       ((sids2c sc (strids@[strid]))
        handle _ => ((c'2c (sids2c' gc (strids@[strid]))) handle _ =>
                     raise Fail ("Unbound structure "^(strLName strlid))))
       (* one fid and lid for one structure toplevel *)
     | StrStr (strd, _) =>
       let val fid = nFid() val lid = nLid()
           val C(sc', ci') = strdec fc gc c strd
       in  C(sc', ci'{fns <- (fid,lid,name)::ci'.fns})
       end
     | SigStr (stre', sige, _) => strexp fc gc c stre' name
     | FctAppStr ((id,_), [stre'], info) =>
       let val (formal,body,tables) = (FC.find fc id)
               handle Not_found => raise Fail ("Unbound functor "^id)
           val actual = strexp fc gc c stre' "@toplevel"
           val fm = ref ([]: (fid * fid) list)
           val lm = ref ([]: (lid * lid) list)
           val nonfn = ref ([]: (fid * lid) list)
           (* for each (f,l,name) \in "formal", *)
           (* find (f',l',name) \in "actual" *)
           (* if name is not a function type, it is dropped *)
           (* and make a substitution {f |-> f', l |-> l'} *)
           (* mkSubst: ctxt' -> ctxt -> unit *)
           fun mkSubst (Ct(gc,ci)) (C(sc,ci')) =
             (L.iter (fn (f,l,n) =>
                         (case f2f n ci'.fns
                            of Some (f',l') => (fm := (f,f')::(!fm);
                                                lm := (l,l')::(!lm))
                             | None => nonfn := (f,l)::(!nonfn))) ci.fns;
              GC.iter (fn id c' =>
                          mkSubst c' ((SC.find sc id) handle Not_found =>
                                    raise Fail ("Unbound structure "^id))) gc)
           val _ = mkSubst formal actual (* \rho_1 = formal |X| actual *)
           (* Dom(\rho_1) \cap Dom(\rho_2) = \emptyset *)
           val (body',(sbf,sbl)) = copyC body (* \rho_1(body) = body *)
           val _ = (fm := sbf@(!fm); lm := sbl@(!lm)) (* \rho_2 o \rho_1 *)
           val _ = addFctApp (getLoc info) (!fm)
           val _ = if !debugflag
                   then (print "\nFctAppStr\nfm=";
                         L.iter (fn (f,f') => (print ("("^(i2s f)^","
                                                      ^(i2s f')^")"))) (!fm);
                         print "\nlm=";
                         L.iter (fn (l,l') => (print ("("^(i2s l)^",:"
                                                      ^(i2s l')^")"))) (!lm);
                         print "\nnon-functions=";
                         L.iter (fn (f,l) => print ("("^(i2s f)^","
                                                    ^(i2s l)^")")) (!nonfn);
                         print "\nbody' start["; prC body'; print "\n]end")
       in  (addTbls (substTbls tables (!fm) (!lm)); body')
       end
     | FctAppStr _ => raise Fail "Functor is not desugared"

(* strdec: fc -> gc -> ctxt -> strdec -> ctxt *)
and strdec fc gc (c as C(sc,ci)) strd =
  case strd
    of SimpleDec (d, _) =>
       let val _ = (cursc := sc; curgc := gc)
           val olf = fid() val oll = lid()
           val env = dec {ftns = ci.fns, owns = owns0} d
           val _ = (cursc := SC.create 0; curgc := GC.create 0)
           val _ = if !debugflag andalso true
                   then (print "\nSimpleDec"; prF ();
                         prC (C(SC.create 0,{ff = olf, lf = fid(),
                                             fl = oll, ll = lid(),
                                            fns = lstMinus env.ftns ci.fns})))
       in  C(SC.create 0,{ff = olf, lf = fid(), fl = oll, ll = lid(),
                          fns = lstMinus env.ftns ci.fns})
       end
     | StrDec (strbs, _) =>
       let val sc' = SC.create 30
           val olf = fid() val oll = lid()
           val _ =  L.iter (fn ((id,_), _, stre, _) =>
                               let val c' = strexp fc gc c stre id
                               in  SC.add sc' id c'
                               end) strbs
           val _ = if !debugflag
                   then (print "\nStrDec";
                         prC (C(sc', {ff = olf, lf = fid(), fl = oll,
                                      ll = lid(), fns = ftns0})))
       in  C(sc', {ff = olf, lf = fid(), fl = oll, ll = lid(),
                   fns = ftns0})
       end
     | SeqStrDec (strds, _) =>
       let val (_, netC) =
           foldl (fn (c1,net) sd => let val c2 = strdec fc gc c1 sd
                                    in  (addC c1 c2, addC net c2)
                                    end)
                 (c, C(SC.create 0, {ff = fid(), lf = fid(), fl = lid(),
                                     ll = lid(), fns = ftns0})) strds
       in  netC
       end

(* spec: gc -> spec -> string -> ctxt' *)
fun spec gc spc name =
  case spc
    of ValSpec (vdscs, _) =>
       let val off = fid() val ofl = lid()
           val ftns =
               foldl (fn fns ((id,_),(_,t),_) =>
                             if isFnty t orelse TyUtil.is_abs t
                             then (nFid(),nLid(),id)::fns else fns) [] vdscs
           val _ = if !debugflag andalso false
                   then (print "\nValSpec";
                         prC' (Ct(GC.create 0,{ff = off, lf = fid(),
                                               fl = ofl, ll = lid(),
                                               fns = ftns})))
       in  Ct(GC.create 0, {ff = off, lf = fid(), fl = ofl, ll = lid(),
                            fns = ftns})
       end
     | IncludeSpec (sige, _) => sigexp gc sige name
     | StrSpec (sdscs, _) =>
       let val off = fid() val ofl = lid()
           val gc' = GC.create 30
           val _ = L.iter (fn ((id,_),sige,_) =>
                              let val c' = sigexp gc sige id
                              in  GC.add gc' id c'
                              end) sdscs
           val _ = if !debugflag andalso true
                   then (print "\nStrSpec";
                         prC' (Ct(gc',{ff = off, lf = fid(),
                                       fl = ofl, ll = lid(), fns = ftns0})))
       in  Ct(gc', {ff = off, lf = fid(), fl = ofl, ll = lid(),
                    fns = ftns0})
       end
     | SeqSpec (spcs, _) => (* refer to (3.38) in "The Definition of nML" *)
       foldl (fn c1 sp => let val c2 = spec gc sp name in addC' c1 c2 end)
             (Ct(GC.create 30,{ff = fid(),lf= fid(),fl = lid(),ll = lid(),
                               fns = ftns0})) spcs
     | _ => Ct(GC.create 0, {ff = fid(),lf = fid(),fl = lid(),ll = lid(),
                             fns = ftns0})

(* sigexp: gc -> sigexp -> string -> ctxt' *)
and sigexp gc sige name =
  case sige
    of VarSig ((id,_), _) =>
       ((copyC' (GC.find gc id))
        handle Not_found => raise Fail ("Unbound signature "^id))
       (* one fid and lid for one signature toplevel *)
     | SigSig (spc, _) =>
       let val fid = nFid() val lid = nLid()
           val Ct(gc', ci') = spec gc spc name
       in  Ct(gc', ci'{fns <- (fid,lid,name)::ci'.fns})
       end
     | _ => raise Fail "strange Sigexp"

(* sigdec: gc -> sigdec -> gc *)
fun sigdec gc (sigbs, _) =
  let val gc' = GC.create 30
  in  (L.iter (fn ((id,_), sige, _) =>
                  let val c' = sigexp gc sige id
                  in GC.add gc' id c' end) sigbs;
       gc')
  end

(* fctdec: basis -> fctdec -> fc *)
fun fctdec b ((fctid,_), [((sigid,_),sige,_)], stre, _) =
    let val arg = sigexp b.gc sige "@toplevel"
        val gc = GC.create 1
        val _ = GC.add gc sigid arg
        val fc = FC.create 1
        val preTbls = getTbls()
        val body = strexp b.fc gc b.c stre "@toplevel"
        val postTbls = getTbls()
        val netTbls = minusTbls postTbls preTbls
        val _ = if !debugflag then (print "\nnetTbls"; prTbls netTbls)
        val _ = setTbls preTbls
        val _ = if !debugflag then (print "\nFctDec"; prC body)
    in  (FC.add fc fctid (arg, body, netTbls); fc)
    end
  | fctdec _ _ = raise Fail "Functor is not desugared"

(* topdec: basis -> topdec -> basis *)
fun topdec b td =
  case td
    of Sig (sigd, _) =>
       {fc = b.fc, gc = addGC b.gc (sigdec b.gc sigd), c = b.c}
     | Fct (fctd, _) =>
       {fc = addFC b.fc (fctdec b fctd), gc = b.gc, c = b.c}
     | Str (strd, _) =>
       {fc = b.fc, gc = b.gc, c = addC b.c (strdec b.fc b.gc b.c strd)}
     | SeqTopDec (topds, _) =>
       let val (_, netB) =
           foldl (fn (b1,net) td => let val b2 = topdec b1 td
                                    in ({fc = addFC b1.fc b2.fc,
                                         gc = addGC b1.gc b2.gc,
                                         c = addC b1.c b2.c},
                                        {fc = addFC net.fc b2.fc,
                                         gc = addGC net.gc b2.gc,
                                         c = addC net.c b2.c})
                                    end) (b,!basis0) topds
       in  netB
       end

(* construct: Absyn.topdec -> basis *)
fun construct absyn = topdec (!basis0) absyn

end (* CfaConstruct *)
