(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

signature CFAFILEIO =
sig
exception Invalid_header
val fileIOext: string
val dump: string -> CfaUtils.basis -> unit
val load: string -> CfaUtils.basis * CfaUtils.tables
end (* CFAFILEIO *)

structure CfaFileIO : CFAFILEIO =
struct

structure CU = CfaUtils
structure L = List

exception Invalid_header

type header = {version: string}
type fileIO = {header: header, basis: CU.basis, tables: CU.tables}

val version = "0.9"
val fileIOext = ".nfa"

(* dump_it: out_channel -> 'a -> unit *)
fun dump_it ch x = Marshal.to_channel ch x []
(* dump_all: out_channel -> fileIO -> unit *)
fun dump_all ch {header = h, basis = b, tables = t} =
  (dump_it ch h; dump_it ch b; dump_it ch t)
(* dump: string -> basis -> unit *)
fun dump file basis =
  let val oc = open_out_bin ((Filename.chop_extension file)^fileIOext)
  in  (dump_all oc {header = {version = version}, basis = basis,
                    tables = CU.getTbls ()};
       close_out oc)
  end

(* load_header: in_channel -> header *)
fun load_header ch =
  let val h = Marshal.from_channel ch
  in  if h.version = version then h else raise Invalid_header
  end
(* load_it: in_channel -> fileIO *)
fun load_it ch = Marshal.from_channel ch
(* load_all: in_channel -> fileIO *)
fun load_all ch =
  let val h = load_header ch val b = load_it ch val t = load_it ch
  in  {header = h, basis = b, tables = t}
  end
(* load: string -> basis * tables *)
fun load file =
  let val ic = open_in_bin file
      val l = (load_all ic) handle x => (close_in ic; raise x)
  in  (close_in ic; (l.basis, l.tables))
  end

end (* CfaFileIO *)
