(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

structure CfaMain =
struct

open Location
open Format

val _ = Utils.cfa := true
val include_dirs = ref ([]: string list)
val include_files = ref ([]: string list)
val prAbsyn = ref false
val prTblF = ref false
val prTime = ref false
val print = print_string
fun set f () = f := true
fun printT phase ftn = if !prTime then Utils.T.time phase ftn else ftn()

fun cfamain filename =
  (if not(Filename.check_suffix filename ".n")
   then raise (Arg.Bad ("don't know what to do with "^filename));
   let val prefixname = Filename.chop_extension filename
       val modulename = Filename.basename prefixname
       val ntyfile = prefixname^".nty"
       val cfafile = prefixname^".nfa"
       val _ = Init.init (!include_dirs @ ["./"])
       fun doLoad() = CfaInit.init (!include_files)
       val _ = printT "0cfaLoad" doLoad
       val ic = open_in_bin filename
       fun doAst() = (input_name := filename;
		      Nparse.implementation_parse (Lexing.from_channel ic))
	             handle x => (close_in ic; raise x)
       val ast = printT "1parse" doAst
       fun doPre() = PreInfer.preinfer ast
       val ast' = printT "2preInfer" doPre
       fun doInf() =
	 let val _ = NEnv.reset_file modulename
	     val absyn = Infer.infer ast'
	 in  (ignore(NEnv.stabilize_file()); Dump.dump ntyfile; absyn)
	 end
       val absyn = printT "3typeInfer" doInf
       fun doPost() = PostInfer.postinfer absyn
       val absyn' = printT "4postInfer" doPost
       val _ = if !prAbsyn then Absyn.pp absyn'
       fun doCfaOne() = CfaConstruct.construct absyn'
       val basis = printT "5cfaConstruct" doCfaOne
       fun doCfaTwo() = CfaSolve.solve ()
       val _ = printT "6cfaSolve" doCfaTwo
       fun doDump() = CfaFileIO.dump cfafile basis
       val _ = printT "7cfaDump" doDump
       val _ = if !prTblF then CfaUtils.prF ()
       (* begin: for Cfa results reporting *)
       val (msg_file, msg_line, msg_chars, msg_to, msg_head) =
           ("",":",".","-", "")
       fun prLoc ppf loc =
         let val (filename, linenum, linebeg) =
                 Linenum.for_position !input_name loc.loc_start
         in  (fprintf ppf "%s%s%s%i" msg_file filename msg_line linenum;
              fprintf ppf "%s%i" msg_chars (loc.loc_start - linebeg);
              fprintf ppf "%s%i" msg_to (loc.loc_end - linebeg);
              fprintf ppf "%s" msg_head)
         end
       val _ = CfaResult.prCfa prLoc
       (* end: for Cfa results reporting *)
       val _ = Utils.T.pp ()
   in  close_in ic
   end)

val optionlist = [
  ("-I",
   Arg.String (fn s => include_dirs := s :: !include_dirs),
   "<dir>  Add <dir> to the list of include directories"),
  ("-F",
   Arg.String (fn s => include_files := s :: !include_files),
   "<file>  Add <file> to the list of previous files"),
  ("-d", Arg.Unit (set Utils.debugflag), " Print debugging information"),
  ("-p", Arg.Unit (set prAbsyn), " Print absyn"),
  ("-f", Arg.Unit (set prTblF), " Print TableF"),
  ("-t", Arg.Unit (set prTime), " Time profile")
]

val usage = "Usage: cfa <options> <files>\nOptions are:"

fun main () =
  (Arg.parse optionlist cfamain usage;
   printf "@.")
  handle x => (printf "@.";
	       (case x
	          of Nlexer.Error(err, _, _) => Nlexer.report_error err
                   | NsyntaxErr.Error err => NsyntaxErr.report_error err
		   | TyError.T err => TyError.pp err
                   | x => raise x);
	       printf "\n")

end (* CfaMain *)

val _ = CfaMain.main ()
