(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

signature CFARESULT =
sig
val getCfaEa1: Location.t -> (CfaUtils.fid * bool * bool) list
val getCfaEa2: CfaUtils.fid -> CfaUtils.fid list
val prCfa: (Format.formatter -> Location.t -> unit) -> unit
val getCfa: Location.t -> (string * Location.t) list
end (* CFARESULT *)

structure CfaResult : CFARESULT =
struct

open Format
open Utils
open CfaUtils

exception GetCfa of fid

(* getCfaEa1: Location.t -> (CfaUtils.fid * bool * bool) list    *)
fun getCfaEa1 loc =
  foldl (fn lst f => case getFids f
                       of Some (_,_,_,_,(b1,b2)) => lst@[(f,b1,b2)]
                        | None => raise GetCfa f) [] (getF(loc2lid loc))
  handle GetCfa f =>
         if f < 0 then [(f,true,true)] else []
       | Fail "loc2lid" => (print "\n\nlocation="; LocationInfo.print loc;
                            prLoc2lid (); raise Fail "loc2lid")

(* getCfaEa2: fid -> fid list *)
fun getCfaEa2 fid = getF (case getFids fid
                            of Some (l,_,_,_,_) => l
                             | None => raise Fail "getCfaEa")

(* prCfa: (Format.formatter -> Location.t -> unit) -> unit *)
fun prCfa prloc =
  List.iter (fn (loc,lid) =>
                let val ftns = CfaUtils.getF lid
                in case ftns
                     of [] => ()
                      | _ =>
                        let val (front, last) = Utils.lstlast ftns []
                            fun prFtns ftn =
                              case CfaUtils.getFids ftn
                                of Some (_,(n::_),l,_,_) =>
                                   (print ("<"^n^", "); prloc std_formatter l;
                                    print_flush(); print ">\n      ")
                                 | _ => ()
                        in  (print "\n("; prloc std_formatter loc;
                             print_flush(); print ")\n    { ";
                             List.iter prFtns front;
                             case CfaUtils.getFids last
                               of Some (_,(n::_),l,_,_) =>
                                  (print ("<"^n^", "); prloc std_formatter l;
                                   print_flush(); print "> }")
                                | _ => ())
                        end
                end) (List.rev(!CfaUtils.result))

(* getCfa: Location.t -> (string * Location.t) list *)
fun getCfa loc =
  foldl (fn lst f => case getFids f
                       of Some (_,(n::_),l,_,_) => lst@[(n,l)]
                        | _ => lst) [] (getF(loc2lid loc))

end (* CfaResult *)
