(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

structure CfaTop =
struct

open Location
open Format
open Lexing
open String_ast.Ast

exception Directive
val _ = Utils.cfa := true
val include_dirs = ref ([]: string list)
val prAbsyn = ref false
val prTblF = ref false
val prTime = ref false
val print = print_string
fun set f () = f := true
fun printT phase ftn = if !prTime then Utils.T.time phase ftn else ftn()

(* Reading function for interactive use *)

val first_line = ref true
val got_eof = ref false

fun refill_lexbuf buffer len =
  if !got_eof then (got_eof := false; 0)
  else (output_string stdout (if !first_line then "# " else "  ");
        flush stdout;
        first_line := false;
        let val i = ref 0
        in  (while !i < len andalso (let val c = input_char stdin
                                     in  String.set buffer (!i) c; c<>'\n'
                                     end)
             do incr i end;
             !i + 1) handle End_of_file => (echo_eof ();
                                            if !i > 0
                                            then (got_eof := true; !i)
                                            else 0)
        end)

fun empty_lexbuf lb =
  let val l = String.length lb.lex_buffer
  in  ignore(lb{lex_abs_pos <- (-l)});
      lb{lex_curr_pos <- l}
  end

(* Discard everything already in a lexer buffer *)

fun loop () =
  let val _ = Init.init ["./"]
      val lb = Lexing.from_function refill_lexbuf
  in  print "nML Control Flow Analyzer, Version 0.9, 08/06/2001, Sukyoung Ryu [ROPAS/KAIST]";
      print_newline();
      input_name := "";
      input_lexbuf := Some lb;
      Sys.catch_break true;
      while true do
      (ignore(empty_lexbuf lb);
       reset();
       first_line := true;
       NEnv.reset_top ();
       let fun doAst() =
             case Nparse.toplevel_parse lb
               of TopDef ast => ast
                | TopExp (es, i) =>
                  Str(SimpleDec(ValDec([],[(WildPat i,SeqExp(es,i),i)],i),i),i)
                | _ => raise Directive
           val ast = printT "1parse" doAst
           fun doPre() = PreInfer.preinfer ast
           val ast' = printT "2preInfer" doPre
           fun doInf() =
	     let val absyn = Infer.infer ast'
	     in  (NEnv.stabilize_top(); absyn)
             end
           val absyn = printT "3typeInfer" doInf
           fun doPost() = PostInfer.postinfer absyn
           val absyn' = printT "4postInfer" doPost
           val _ = if !prAbsyn then Absyn.pp absyn'
           fun doCfaOne() = CfaConstruct.construct absyn'
           val basis = printT "5cfaConstruct" doCfaOne
           fun doCfaTwo() = CfaSolve.solve ()
           val _ = printT "6cfaSolve" doCfaTwo
           val _ = if !prTblF then CfaUtils.prF ()
           (* begin: for Cfa results reporting *)
           fun prLoc ppf loc = fprintf ppf "%i-%i" loc.loc_start loc.loc_end
           val _ = CfaResult.prCfa prLoc
           (* end: for Cfa results reporting *)
       in  Utils.T.pp ()
       end) handle x => (printf "@.";
	                 (case x
	                    of Nlexer.Error(err, _, _) =>
                               Nlexer.report_error err
                             | NsyntaxErr.Error err =>
                               NsyntaxErr.report_error err
		             | TyError.T err => TyError.pp err
                             | Directive => ()
                             | x => raise x);
	                 printf "\n")
      end  
  end

fun top () = loop ()

end (* CfaTop *)

val _ = (CfaTop.top ()) handle End_of_file => ()
