(*
 Sukyoung Ryu, The LET Project, KAIST

 Copyright(c) 2001 Research On Program Analysis System
 National Creative Research Initiative Center
 Korea Advanced Institute of Science & Technology
 http://ropas.kaist.ac.kr

 All rights reserved. This file is distributed under the terms of
 an Open Source License.
*)

signature CFAUTILS =
sig

(* fid: function id **************************************************)
type fid = int
val fid: unit -> fid
val nFid: unit -> fid
val mkFids: int -> unit
val mkFids': int -> fid -> fid list -> fid list

(* lid: lexp id ******************************************************)
type lid = int
val lid: unit -> lid
val nLid: unit -> lid
val mkLids: int -> lid list -> lid list

val result: ((Location.t * lid) list) ref

(* lexp **************************************************************)
type lexp = Ve of Absyn.varlongid * fid * int (* owner, pos *)
          | Ae of lid                (* lid1 *)
          | FAe of lid               (* lid1 *)
          | De of Ty.ty              (* decon by type *)

(* environment *******************************************************)
type ftn = fid * lid * string (* fid, lid, fname *)
type ftn_env = ftn list
type own = fid * string * int (* owner, varname, reccnt *)
type own_env = own list
type env = {ftns: ftn_env, owns: own_env}
type tables = {fids: (lid*string list*Location.t*int*(bool*bool)) Utils.IM.t,
               tblF: Utils.IS.t Utils.IM.t,
               tblT: Utils.IS.t Utils.TM.t,
               tblLa: Utils.IS.t Utils.IM.t,
               tblLb: fid Utils.IM.t,
               tblE: Utils.IS.t Utils.IM.t,
               tblAb: (lid list * Ty.ty) list Utils.IM.t,
               tblApAb: lid list list Utils.IM.t,
               tblApL: lid Utils.IM.t,
               tblLid2le: lexp Utils.IM.t,
               tblLoc2lid: lid Utils.LM.t,
               tblFctApp: (fid * fid) list Utils.LM.t}
type cfainfo = {ff: fid, lf: fid, fl: lid, ll: lid, fns: ftn_env}
structure SC : H where type key = StringInfo.stridinfo
type sc = ctxt SC.t
and ctxt = C of sc * cfainfo
structure GC : H where type key = StringInfo.sigidinfo
type gc = ctxt' GC.t
and ctxt' = Ct of gc * cfainfo
structure FC : H where type key = StringInfo.fctidinfo
type fc = (ctxt' * ctxt * tables) FC.t
type basis = {fc: fc, gc: gc, c: ctxt}

val ftns0: ftn_env
val owns0: own_env
val initftns: ftn_env ref
val env0: env
val ci0: cfainfo
val basis0: basis ref
val cursc: sc ref
val curgc: gc ref
val b2fc: basis -> fc
val b2gc: basis -> gc
val ci2ff: cfainfo -> fid
val ci2lf: cfainfo -> fid

val c'2c: ctxt' -> ctxt
val sids2c: sc -> Absyn.strid list -> ctxt
val sids2c': gc -> Absyn.strid list -> ctxt'
val addFC: fc -> fc -> fc
val addGC: gc -> gc -> gc
val addSC: sc -> sc -> sc
val addCI: cfainfo -> cfainfo -> cfainfo
val addC: ctxt -> ctxt -> ctxt
val addC': ctxt' -> ctxt' -> ctxt'
val copyC: ctxt -> ctxt * ((fid * fid) list * (lid * lid) list)
val copyC': ctxt' -> ctxt'

val fPops: ftn_env -> ftn_env -> ftn_env
val fPush: ftn -> ftn_env -> ftn_env
val fPushs: ftn_env -> ftn_env -> ftn_env
val oPops: own_env -> own_env -> own_env
val oPush: own -> own_env -> own_env
val oPushs: own_env -> own_env -> own_env
val f2f: string -> ftn_env -> (fid * lid) option
val fnm2fid: Absyn.varlongid -> ftn_env -> (fid * lid) option
val getOwner: string -> own_env -> int * int
val mkOwn: fid -> int -> Absyn.pat -> own_env

(* tables ************************************************************)
val count: int ref
val flag: bool ref

val addFids: fid * lid * string * Location.t * int * (bool * bool) -> unit
val addFids': fid * lid * string list * Location.t * int * (bool * bool)
              -> unit
val getFids:
    fid -> (lid * string list * Location.t * int * (bool * bool)) option
val getAri: fid -> int

val addF: lid -> fid -> unit
val getF: lid -> fid list
val addT: Ty.ty -> fid -> unit
val addLa: fid -> lid list -> unit
val getLa: fid -> lid list
val addLb: lid -> fid -> unit
val addE: lid list -> bool -> lid
val addAb: lid -> fid -> int -> Ty.ty -> unit
val addApAb: lid -> lid list -> unit
val addApL: lid -> lid -> unit
val addDmAp: lid -> unit
val lid2le: lid -> lexp -> unit
val getLid2le: lid -> lexp option
val addLoc2lid: Location.t -> lid -> unit
val loc2lid: Location.t -> lid
val addFctApp: Location.t -> (fid * fid) list -> unit
val getFctApp: Location.t -> (fid * fid) list

val prC: ctxt -> unit
val prC': ctxt' -> unit
val prLoc2lid: unit -> unit
val prF: unit -> unit

val getHifn: lid -> ftn_env -> fid * fid
val fidsbytys: unit -> unit
val getTbls: unit -> tables
val setTbls: tables -> unit
val addTbls: tables -> unit
val minusTbls: tables -> tables -> tables
val substTbls: tables -> (fid * fid) list -> (lid * lid) list -> tables
val prTbls: tables -> unit
val initCfa: unit -> unit
val initCfaConstruct: basis * tables -> unit
val initCfaSolve: unit -> unit

val cfasE: unit -> unit
val cfasAp: unit -> unit
val cfasLb: unit -> unit

end (* CFAUTILS *)

structure CfaUtils : CFAUTILS =
struct

open Utils
structure L = List
structure A = Absyn
structure AU = AbsynUtils

(* fid: function id **************************************************)
type fid = int
val idf = ref 0 (* function id *)
(* fid: unit -> fid *)
fun fid () = !idf
(* nFid: unit -> fid *)
fun nFid () = (inc idf; !idf)
(* mkFids: int -> unit *)
(* mkFids cnt = (ofid+1;ofid+2;...;ofid+cnt;()) *)
fun mkFids cnt = if cnt=0 then () else (ignore(nFid()); mkFids (cnt-1))
(* mkFids': int -> fid -> fid list -> fid list *)
(* mkFids' cnt ofid [] = [ofid+1,ofid+2,...,ofid+cnt] *)
fun mkFids' cnt fid res =
  if cnt=0 then res else mkFids' (cnt-1) (fid+1) (res@[fid+1])

(* lid: lexp id ******************************************************)
type lid = int
val idl = ref 0 (* lexp id *)
(* lid: unit -> lid *)
fun lid () = !idl
(* nLid: unit -> lid *)
fun nLid () = (inc idl; !idl)
(* mkLids: int -> lid list -> lid list *)
(* mkLids cnt [] = [olid+cnt,...,olid+2,olid+1] *)
fun mkLids cnt res = if cnt=0 then res else mkLids (cnt-1) ((nLid())::res)

val result = ref ([]: (Location.t * lid) list)

(* lexp **************************************************************)
(* lid |-> Ve (name,own,pos) : tblA, tblA' *)
(* lid |-> Ae lid1           : tblL *)
(* lid |-> FAe lid1 *)
(* lid |-> De ty             : lid >= li where type(li) =unif ty *)
type lexp = Ve of Absyn.varlongid * fid * int (* owner, pos *)
          | Ae of lid                (* lid1 *)
          | FAe of lid               (* lid1 *)
          | De of Ty.ty              (* decon by type *)

(* environment *******************************************************)
type ftn = fid * lid * string (* fid, lid, fname (for debugging) *)
type ftn_env = ftn list
type own = fid * string * int (* owner, varname, reccnt *)
type own_env = own list
type env = {ftns: ftn_env, owns: own_env}
type tables = {fids: (lid*string list*Location.t*int*(bool*bool)) Utils.IM.t,
               tblF: Utils.IS.t Utils.IM.t,
               tblT: Utils.IS.t Utils.TM.t,
               tblLa: Utils.IS.t Utils.IM.t,
               tblLb: fid Utils.IM.t,
               tblE: Utils.IS.t Utils.IM.t,
               tblAb: (lid list * Ty.ty) list Utils.IM.t,
               tblApAb: lid list list Utils.IM.t,
               tblApL: lid Utils.IM.t,
               tblLid2le: lexp Utils.IM.t,
               tblLoc2lid: lid Utils.LM.t,
               tblFctApp: (fid * fid) list Utils.LM.t}
type cfainfo = {ff: fid, lf: fid, fl: lid, ll: lid, fns: ftn_env}
structure SC = MakeHash (struct type t = StringInfo.stridinfo val equal = (=)
                                val hash = Hashtbl.hash end)
type sc = ctxt SC.t
and ctxt = C of sc * cfainfo
structure GC = MakeHash (struct type t = StringInfo.sigidinfo val equal = (=)
                                val hash = Hashtbl.hash end)
type gc = ctxt' GC.t
and ctxt' = Ct of gc * cfainfo
structure FC = MakeHash (struct type t = StringInfo.fctidinfo val equal = (=)
                                val hash = Hashtbl.hash end)
type fc = (ctxt' * ctxt * tables) FC.t
type basis = {fc: fc, gc: gc, c: ctxt}

val ftns0 = nil
val owns0 = nil
val initftns = ref ftns0
val env0 = {ftns = !initftns, owns = owns0}: env
val ci0 = {ff= 0,lf= 0,fl= 0,ll= 0,fns= ftns0}
val basis0 = ref {fc = FC.create 30, gc = GC.create 30,
                  c = C(SC.create 30, ci0)}
val cursc = ref (SC.create 0)
val curgc = ref (GC.create 0)
(* b2fc: basis -> fc *)
fun b2fc b = b.fc
(* b2gc: basis -> gc *)
fun b2gc b = b.gc
(* ci2ff: cfainfo -> fid *)
fun ci2ff ci = ci.ff
(* ci2lf: cfainfo -> fid *)
fun ci2lf ci = ci.lf

(* prci: cfainfo -> unit *)
fun prci {ff, lf, fl, ll, fns} =
  (print ("\n        cfainfo:ff="^(i2s ff)^" lf="^(i2s lf)^" fl="^(i2s fl)
          ^" ll="^(i2s ll)^" fns=");
   L.iter (fn (f,l,n) => print ("<"^(i2s f)^","^(i2s l)^","^n^">")) fns)

(* prC: ctxt -> unit *)
fun prC (c as C(sc,ci)) =
  let val cs = ref ([]: ctxt list)
      val scs = ref ([]: sc list)
      fun subprC (c' as C(sc',ci')) =
        if L.mem c' (!cs) then print "cycle"
        else (print "\nContext:"; cs := c'::(!cs); prSC sc'; prci ci')
      (* prSC: sc -> unit *)
      and prSC sc' =
        if L.mem sc' (!scs) then print "cycle"
        else (print "sc:"; scs := sc'::(!scs);
              SC.iter (fn strid c' => (print (strid^"="); subprC c')) sc')
  in  (cs := nil; scs := nil; 
       print "\nContext:"; cs := c::(!cs); prSC sc; prci ci)
  end

(* prC': ctxt' -> unit *)
fun prC' (c as Ct(gc,ci)) =
  let val c's = ref ([]: ctxt' list)
      val gcs = ref ([]: gc list)
      fun subprC' (c' as Ct(gc',ci')) =
        if L.mem c' (!c's) then print "cycle"
        else (print "\nContext:"; c's := c'::(!c's); prGC gc'; prci ci')
      (* prGC: gc -> unit *)
      and prGC gc' =
        if L.mem gc' (!gcs) then print "cycle"
        else (print "gc:"; gcs := gc'::(!gcs);
              GC.iter (fn sigid c' => (print (sigid^"="); subprC' c')) gc')
  in  (c's := nil; gcs := nil;
       print "\nContext:"; c's := c::(!c's); prGC gc; prci ci)
  end

(* c'2c: ctxt' -> ctxt *)
fun c'2c (Ct(gc,ci)) = C(gc2sc gc,ci)
(* gc2sc: gc -> sc *)
and gc2sc gc =
  let val sc = SC.create 30
  in  (GC.iter (fn sigid c' => SC.add sc sigid (c'2c c')) gc; sc)
  end

(* sids2c: sc -> strid list -> ctxt *)
fun sids2c sc [] = raise Fail "strange strid list"
  | sids2c sc [(strid,_)] = ((SC.find sc strid)
                             handle Not_found => raise Fail ("sids2: "^strid))
  | sids2c sc strids =
    let val (front, (last,_)) = lstlast strids []
        val C(sc'',_) =
            foldl (fn (C(sc',_)) (id,_) => ((SC.find sc' id)
                       handle Not_found => raise Fail ("sids2: "^id)))
                  (C(sc,ci0)) front
    in  (SC.find sc'' last) handle Not_found => raise Fail ("sids2: "^last)
    end

(* sids2c': gc -> strid list -> ctxt' *)
fun sids2c' gc [] = raise Fail "strange strid list"
  | sids2c' gc [(sigid,_)] = ((GC.find gc sigid)
                             handle Not_found => raise Fail("sids2': "^sigid))
  | sids2c' gc strids =
    let val (front, (last,_)) = lstlast strids []
        val Ct(gc'',_) =
            foldl (fn (Ct(gc',_)) (id,_) => ((GC.find gc' id)
                       handle Not_found => raise Fail ("sids2': "^id)))
                  (Ct(gc,ci0)) front
    in  (GC.find gc'' last) handle Not_found => raise Fail ("sids2': "^last)
    end

(* addFC: fc -> fc -> fc *)
fun addFC fc1 fc2 = (FC.iter (fn fctid c => FC.add fc1 fctid c) fc2; fc1)

(* addGC: gc -> gc -> gc *)
fun addGC gc1 gc2 = (GC.iter (fn sigid c => GC.add gc1 sigid c) gc2; gc1)

(* addSC: sc -> sc -> sc *)
fun addSC sc1 sc2 = (SC.iter (fn strid c => SC.add sc1 strid c) sc2; sc1)

(* addCI: cfainfo -> cfainfo -> cfainfo *)
fun addCI ci1 ci2 =
  {ff = ci1.ff, lf = ci2.lf, fl = ci1.fl, ll = ci2.ll, fns = ci2.fns@ci1.fns}

(* addC: ctxt -> ctxt -> ctxt *)
fun addC (C(sc1,ci1)) (C(sc2,ci2)) = C(addSC sc1 sc2, addCI ci1 ci2)

(* addC': ctxt' -> ctxt' -> ctxt' *)
fun addC' (Ct(gc1,ci1)) (Ct(gc2,ci2)) = Ct(addGC gc1 gc2, addCI ci1 ci2)

(* copyCI: cfainfo -> cfainfo * ((fid * fid) list * (lid * lid) list) *)
fun copyCI ci =
  let fun mkSubst init offset count =
        let fun f 0 res = (init, init+offset)::res
              | f c res = f (c-1) ((init+c, init+c+offset)::res)
        in  f (count-1) []
        end
      val (nff,nlf,substf) =
          if ci.ff > ci.lf then (fid()+1,fid(),[])
          else let val ff = fid()+1
                   val count = ci.lf-ci.ff+1
                   val _ = mkFids count
               in  (ff, fid(), mkSubst ci.ff (ff-ci.ff) count) end
      val (nfl,nll,substl) =
          if ci.fl > ci.ll then (lid()+1,lid(),[])
          else let val fl = lid()+1
                   val count = ci.ll-ci.fl+1
                   val _ = mkLids count []
               in  (fl, lid(), mkSubst ci.fl (fl-ci.fl) count) end
      val nfns = L.map (fn (f,l,n) => (f-ci.ff+nff,l-ci.fl+nfl,n)) ci.fns
  in  ({ff = nff, lf = nlf, fl = nfl, ll = nll, fns = nfns}, (substf,substl))
  end

(* copyGC: gc -> gc *)
fun copyGC gc = 
  let val gc' = GC.create 10
  in  (GC.iter (fn sigid c => GC.add gc' sigid (copyC' c)) gc; gc')
  end

(* copyC': ctxt' -> ctxt' *)
and copyC' (Ct(gc,ci)) =
  let val (ci',_) = copyCI ci in Ct(copyGC gc, ci') end

(* copySC: sc -> sc * ((fid * fid) list * (lid * lid) list) *)
fun copySC sc =
  let val sc' = SC.create 10
      val substf = ref ([]: (fid * fid) list)
      val substl = ref ([]: (lid * lid) list)
  in  (SC.iter (fn strid c =>
                   let val (newc, (sbf,sbl)) = copyC c
                   in  (SC.add sc' strid newc;
                        substf := sbf@(!substf); substl := sbl@(!substl))
                   end) sc; (sc', (!substf,!substl)))
  end

(* copyC: ctxt -> ctxt * ((fid * fid) list * (lid * lid) list) *)
and copyC (C(sc,ci)) =
  let val (ci',(sbf1,sbl1)) = copyCI ci
      val (sc',(sbf2,sbl2)) = copySC sc
  in  (C(sc', ci'), (sbf1@sbf2,sbl1@sbl2))
  end

(* fPops: ftn_env -> ftn_env -> ftn_env *)
fun fPops ftns1 ftns2 = lstMinus ftns1 ftns2
(* fPush: ftn -> ftn_env -> ftn_env *)
fun fPush ftn ftns = ftn::ftns
(* fPushs: ftn_env -> ftn_env -> ftn_env *)
fun fPushs ftns1 ftns2 = ftns1@ftns2

(* oPops: own_env -> own_env -> own_env *)
fun oPops owns1 owns2 = lstMinus owns1 owns2
(* oPush: own -> own_env -> own_env *)
fun oPush own owns = own::owns
(* oPushs: own_env -> own_env -> own_env *)
fun oPushs owns1 owns2 = owns1@owns2

(* f2f: string -> ftn_env -> (fid * lid) option *)
fun f2f fnm nil = None
  | f2f fnm ((fid,lid,fname)::rest) =
    if fnm = fname then Some (fid,lid) else f2f fnm rest
(* fnm2fid: Absyn.varlongid -> ftn_env -> (fid * lid) option *)
fun fnm2fid ([], (fnm,_), _) ftns =
    (case f2f fnm ftns
       of Some r => Some r                                (* local function *)
        | None => (case Perv.getPrimitives fnm        (* primitive function *)
                     of Some fid => Some (fid,-1) | None => None))
  | fnm2fid (vlid as (strids, (fnm,_), _)) ftns =
    (let val C(_,ci) = sids2c (!cursc) strids       (* preanalyzed function *)
     in  f2f fnm ci.fns
     end handle Fail _ =>
         (let val Ct(_,ci) = sids2c' (!curgc) strids    (* functor argument *)
          in  f2f fnm ci.fns
          end handle Fail _ =>                        (* pervasive function *)
              (case Perv.getPervasives (AU.varLName vlid)
                 of Some fid => Some (fid,-1)
                  | None => raise Fail ("fnm2fid:"^(AU.varLName vlid)))))

(* getOwner: string -> own_env -> int * int *)
fun getOwner v nil = raise Fail ("getOwner:"^v)
  | getOwner v ((f,v',n)::rest) = if v=v' then (f,n) else getOwner v rest

(* mkOwn: fid -> int -> pat -> own_env *)
fun mkOwn fid cnt pat =
  case pat
    of A.VarPat (vid,t,_) =>
       if AU.isFnty t then oPush (fid,AU.varName vid,cnt) owns0 else owns0
     | A.AppPat (_,p,_,_) => mkOwn fid (-1) p
     | A.RecordPat (fields,_,_) =>
       foldl (fn res (_,p,_) => oPushs res (mkOwn fid (-1) p)) owns0 fields
     | A.SubRecordPat (fields,_,_) =>
       foldl (fn res (_,p,_) => oPushs res (mkOwn fid (-1) p)) owns0 fields
     | A.RefPat (p,_,_) => mkOwn fid (-1) p
     | A.AsPat (vid,p,t,_) =>
       if AU.isFnty t then oPush (fid,AU.varName vid,cnt) (mkOwn fid cnt p)
       else mkOwn fid cnt p
     | A.TuplePat (pats,_,_) =>
       if cnt = 0
       then snd (foldl (fn (c,res) p => (c+1,oPushs res (mkOwn fid c p)))
                         (1, owns0) pats)
       else foldl (fn res p => oPushs res (mkOwn fid (-1) p)) owns0 pats
     | A.ListPat (pats,_,_) =>
       foldl (fn res p => oPushs res (mkOwn fid (-1) p)) owns0 pats
     | A.ArrayPat (pats,_,_) =>
       foldl (fn res p => oPushs res (mkOwn fid (-1) p)) owns0 pats
     | A.OrPat (pats,_,_) =>
       foldl (fn res p => oPushs res (mkOwn fid (-1) p)) owns0 pats
     | _ => owns0

(* tables ************************************************************)
(* to dump: ftns, fids, tblF, tblT, lid2le, tblLa, tblLb, tblE *)

val count = ref 0     (* the number of computations *)
val flag = ref true (* if true then need to compute cfa again *)

(* fids *********************************************************)
(* fname list: because of AsPat (val a as b = fn x => x) *)
(* fid, lid, name list, loc, number of arguments, (input type, output type) *)
val fids = ref IM.empty
(* addFids: fid * lid * string * Location.t * int * (bool * bool) -> unit *)
fun addFids (fid, lid, name, loc, arg, (t1,t2)) =
  fids := IM.add fid (lid, [name], loc, arg, (t1,t2)) (!fids)
(* addFids': fid * lid * string list * Location.t * int * (bool * bool) *)
(*        -> unit *)
fun addFids' (fid, lid, names, loc, arg, (t1,t2)) =
  fids := IM.add fid (lid, names, loc, arg, (t1,t2)) (!fids)
(* getFids: *)
(* fid -> (lid * string list * Location.t * int * (bool * bool)) option *)
fun getFids fid = (Some (IM.find fid (!fids))) handle Not_found => None
(* getAri: fid -> int *)
(* get the arity of fid *)
fun getAri fid =
if fid < 0 then Perv.getAri fid
else (case IM.find fid (!fids) of (_,_,_,a,_) => a)
     handle Not_found => -1 (* raise Fail "getAri" *)

(* tblF *********************************************************)
(* lid |-> fid set *) (* l >= f *)
val tblF = ref IM.empty
(* getF0: lid -> fid set *)
fun getF0 lid = IM.find lid (!tblF) handle Not_found => IS.empty
(* getF: lid -> fid list *)
fun getF lid = IS.elements (getF0 lid)
(* addF: lid -> fid -> unit *)
fun addF lid fid = tblF := IM.add lid (IS.add fid (getF0 lid)) (!tblF)
(* addF': lid -> fid -> unit *)
fun addF' lid fid =
  let val prev = getF0 lid
  in  if  not(IS.mem fid prev)
      then (tblF := IM.add lid (IS.add fid prev) (!tblF); flag := true)
  end
(* setF: lid -> fid list -> unit *)
fun setF lid fids = tblF := IM.add lid (list2IS fids) (!tblF)

(* tblT *********************************************************)
(* ty |-> fid set *)
val tblT = ref TM.empty
(* getT0: ty -> fid set *)
fun getT0 ty = TM.find ty (!tblT) handle Not_found => IS.empty
(* addT: ty -> fid -> unit *)
fun addT ty fid =
  let val t = AU.clearTy ty
  in  tblT := TM.add t (IS.add fid (getT0 t)) (!tblT)
  end

(* tblLa ********************************************************)
(* fid |-> lid(ftn typed body) set *)
val tblLa = ref IM.empty
(* getLa0: fid -> lid set *)
fun getLa0 fid = IM.find fid (!tblLa) handle Not_found => IS.empty
(* getLa: fid -> lid list *)
fun getLa fid = IS.elements (getLa0 fid)
(* addLa: fid -> lid list -> unit *)
fun addLa fid (bodies:lid list) =
  tblLa := IM.add fid (IS.union (list2IS bodies) (getLa0 fid)) (!tblLa)
(* addLa': fid -> lid set -> unit *)
fun addLa' fid lids =
  let val prev = getLa0 fid
  in  if  not(IS.subset lids prev)
      then (tblLa := IM.add fid (IS.union lids prev) (!tblLa);
            flag := true)
  end

(* tblLb ********************************************************)
(* for functions whose bodies collected later *)
(* val a = ... where a is a function type *)
(* record (lid, fid) of a to tblLb *)
(* and then when lid |-> fid' is added to tblF *)
(* add fid |-> body of fid' to tblLa (by addF') *)
(* lid |-> fid *)
val tblLb = ref IM.empty
(* addLb: lid -> fid -> unit *)
fun addLb lid fid = tblLb := IM.add lid fid (!tblLb)

(* tblE *********************************************************)
(* lid |-> lid set *) (* l >= l' *)
val tblE = ref IM.empty
(* getE0: lid -> lid set *)
fun getE0 lid = IM.find lid (!tblE) handle Not_found => IS.empty
(* getE: lid -> lid list *)
fun getE lid = IS.elements (getE0 lid)
(* addE0: lid -> lid -> unit *)
fun addE0 lid lid' = tblE := IM.add lid (IS.add lid' (getE0 lid)) (!tblE)
(* addE': lid -> lid -> unit *)
fun addE' lid lid' =
  let val prev = getE0 lid
  in  if  not(IS.mem lid' prev)
      then (tblE := IM.add lid (IS.add lid' prev) (!tblE); flag := true)
  end

(* addE: lid list -> bool -> lid *)
fun addE [] isExnty = if isExnty then -1 else raise Fail "addE"
  | addE lids _ = let val lid = nLid() in (L.iter (addE0 lid) lids; lid) end

(* tblAb ********************************************************)
(* fid |-> (lid list * ty) list *) (* l >= \lambda{li} *)
val tblAb = ref IM.empty
(* addAb: lid -> fid -> int -> ty -> unit *)
fun addAb lid owner pos t =
  let fun mkNil 0 = []
        | mkNil 1 = [([],Ty.type_unit)]
        | mkNil i = ([],Ty.type_unit)::(mkNil (i-1))
      fun addArg [] 0 lid = [([lid],t)]
        | addArg [(lids,_)] 0 lid = [(lid::lids,t)]
        | addArg _ 0 lid = raise Fail "addArg"
        | addArg args pos lid =
          let val len = L.length args
          in  if len >= pos
              then let val (middle,_) = L.nth args (pos-1)
                   in  (take (args,pos-1))@[(lid::middle,t)]@(drop (args,pos))
                   end
              else args@(mkNil (pos-len-1))@[([lid],t)]
          end
  in  (case IM.find owner (!tblAb)
         of args => tblAb := IM.add owner (addArg args pos lid) (!tblAb))
      handle Not_found =>
             tblAb := IM.add owner (addArg nil pos lid) (!tblAb)
  end

(* tblApAb ******************************************************)
(* previous tblA & tblA'*****************************************)
(* lid |-> (lid list) list *) (* _ >= @l'.{li} where hasFnty {li} *)
val tblApAb = ref IM.empty
(* addApAb: lid -> lid list -> unit *)
fun addApAb lid1 lids =
  let fun merge [] [] = []
        | merge [] l = L.map (fn i => [i]) l
        | merge l [] = l
        | merge (h1::t1) (h2::t2) =
          (if L.mem h2 h1 then h1 else h2::h1)::(merge t1 t2)
  in  (case IM.find lid1 (!tblApAb)
         of args => tblApAb := IM.add lid1 (merge args lids) (!tblApAb))
      handle Not_found =>
             tblApAb := IM.add lid1 (L.map (fn i => [i]) lids) (!tblApAb)
  end

(* tblApL *******************************************************)
(* lid |-> lid *) (* l >= @l'._ where isFnty (l) *)
val tblApL = ref IM.empty
(* addApL: lid -> lid -> unit *)
fun addApL lid lid1 = tblApL := IM.add lid lid1 (!tblApL)

(* applid whose type is not a function type *)
val dummyApp = ref (nil: int list)
(* addDmAp: lid -> unit *)
fun addDmAp lid = dummyApp := lid::(!dummyApp)

(* tblLid2le ****************************************************)
(* lid |-> lexp *)
val tblLid2le = ref IM.empty
(* lid2le: lid -> lexp -> unit *)
fun lid2le lid le = tblLid2le := IM.add lid le (!tblLid2le)
(* getLid2le: lid -> lexp option *)
fun getLid2le lid = (Some (IM.find lid (!tblLid2le))) handle Not_found => None

(* tblLoc2lid ***************************************************)
(* location |-> lid *)
val tblLoc2lid = ref LM.empty
(* addLoc2lid: Location.t -> lid -> unit *)
fun addLoc2lid loc lid = tblLoc2lid := LM.add loc lid (!tblLoc2lid)
(* loc2lid: Location.t -> lid *)
fun loc2lid loc =
  (LM.find loc (!tblLoc2lid)) handle Not_found => raise Fail "loc2lid"

(* tblFctApp ****************************************************)
(* fid substitutions in functor applications *)
(* location |-> (fid * fid) list *)
val tblFctApp = ref LM.empty
(* addFctApp: Location.t -> (fid * fid) list -> unit *)
fun addFctApp loc subst = tblFctApp := LM.add loc subst (!tblFctApp)
(* getFctApp: Location.t -> (fid * fid) list *)
fun getFctApp loc =
  (LM.find loc (!tblFctApp)) handle Not_found => raise Fail "getFctApp"

(* table utils **************************************************)
    (* prFids: fids -> unit *)
    fun prFids fds =
      (print "\n[Fids]";
       IM.iter (fn fid (l,ns,_,a,_) =>
                   print ("\nfid:"^(i2s fid)^" lid:"^(i2s l)^" arg:"^(i2s a)^
                          " "^(L.hd ns)))
               (!fds);
       print "\n")

    (* prTblIL: IM.t ref -> string -> unit *)
    fun prTblIL tbl name =
      (print ("\n["^name^"]");
       IM.iter (fn id ids =>
                   (print ("\nid:"^(i2s id)^" ids:");
                    L.iter (fn is =>
                               (print "[";
                                L.iter (fn i => print (" "^(i2s i))) is;
                                print "]")) ids))
               (!tbl); print "\n"; Format.print_flush())

    (* prTblIS: IM.t ref -> string -> unit *)
    fun prTblIS tbl name =
      (print ("\n["^name^"]");
       IM.iter (fn id ids => (print ("\nid:"^(i2s id)^" ids:");
                              IS.iter (fn i => print(" "^(i2s i))) ids))
               (!tbl); print "\n"; Format.print_flush())

    (* prTblI: IM.t ref -> string -> unit *)
    fun prTblI tbl name =
      (print ("\n["^name^"]");
       IM.iter (fn id id' => (print ("\nid:"^(i2s id)^" id':"^(i2s id'))))
               (!tbl); print "\n"; Format.print_flush())

    (* prTblAb: tblAb -> unit *)
    fun prTblAb tblab =
      (print "\n[tblAb]";
       IM.iter (fn lid args =>
                   (print ("\nfid:"^(i2s lid)^" args:");
                    L.iter (fn (lids,_) =>
                               (print "[";
                                L.iter (fn l => print ((i2s l)^" ")) lids;
                                print "]")) args))
               (!tblab); print "\n"; Format.print_flush())

    (* prTblE: unit -> unit *)
    fun prTblE () = prTblIS tblE "tblE"

    (* printLe: lexp -> unit *)
    fun printLe le =
      case le
        of Ve (name, own, pos) =>
           print ("Ve (name:"^(AU.varLName name)^" owner:"^(i2s own)
                  ^" pos:"^(i2s pos)^")")
         | Ae lid1 => print ("\n Ae lid1:"^(i2s lid1))
         | FAe lid1 => print ("\n FAe lid1:"^(i2s lid1))
         | De ty => print "De "

    (* printLexp: tblLid2le -> unit *)
    fun printLexp lid2le =
      (print "\n[Lexp]";
       IM.iter (fn lid le => (print ("\nlid="^(i2s lid)); printLe le))
               (!lid2le);
       print "\n"; Format.print_flush())

(* prLoc2I: LM.t ref -> string -> unit *)
fun prLoc2I tbl name =
  (print ("\n["^name^"]");
   LM.iter (fn loc id => (print "\nloc:";
                          Location.print Format.std_formatter loc;
                          print (" id:"^(i2s id))))
           (!tbl); print "\n"; Format.print_flush())

(* prLoc2lid: unit -> unit *)
fun prLoc2lid () = prLoc2I tblLoc2lid "tblLoc2lid"

(* prF: unit -> unit *)
fun prF () =
  (prFids fids; prTblIS tblF "tblF";
   prTblIS tblLa "tblLa"; prTblI tblLb "tblLb";
   prTblIS tblE "tblE"; prTblAb tblAb;
   prTblIL tblApAb "tblApAb"; prTblI tblApL "tblApL";
   printLexp tblLid2le; prLoc2I tblLoc2lid "tblLoc2lid")

(* getHifn: lid -> ftn_env -> fid * fid *)
fun getHifn lid ftns =      (* get the higher order function id *)
  case getF lid             (* the higher order fid, original fid *)
    of [fid] => (fid, fid)
     | _ => (case getLid2le lid
               of Some (Ve (name,_,_)) =>
                  (case fnm2fid name ftns
                     of Some (fid,_) => (fid,fid) | None => (-1,-1))
                | Some (Ae lid1) =>
                  let val (fid,org) = getHifn lid1 ftns
                  in  if org <> -1 then (fid+1,org) else (fid,org)
                  end
                | _ => (-1,-1))

(* ty |-> fid set *)
(* cache of tblT *)
val t2fT = ref TM.empty
(* get fids whose types unify with 'ty' *)
(* ty2fids: ty -> fid list *)
fun ty2fids ty =
  let val ty' = AU.clearTy ty
  in  IS.elements (TM.find ty' (!t2fT))
      handle Not_found =>
             let val fids = TM.fold (fn t fds res =>
                                        if AU.unifiable ty' t
                                        then IS.union fds res
                                        else res) (!tblT) IS.empty
             in  (t2fT := TM.add ty' fids (!t2fT); IS.elements fids)
             end
  end

(* fidsbytys: unit -> unit *)
fun fidsbytys () =
  IM.iter (fn lid lexp => (case lexp of De ty => setF lid (ty2fids ty)
                                      | _ => ())) (!tblLid2le)

(* getTbls: unit -> tables *)
fun getTbls () =
  {fids = !fids, tblF = !tblF, tblT = !tblT, tblLa = !tblLa, tblLb = !tblLb,
   tblE = !tblE, tblAb = !tblAb, tblApAb = !tblApAb, tblApL = !tblApL,
   tblLid2le = !tblLid2le, tblLoc2lid = !tblLoc2lid, tblFctApp = !tblFctApp}

(* except tblLoc2lid *)
(* setTbls: tables -> unit *)
fun setTbls t =
  (fids := t.fids; tblF := t.tblF; tblT := t.tblT;
   tblLa := t.tblLa; tblLb := t.tblLb; tblE := t.tblE;
   tblAb := t.tblAb; tblApAb := t.tblApAb; tblApL := t.tblApL;
   tblLid2le := t.tblLid2le; (* tblLoc2lid := t.tblLoc2lid; *)
   tblFctApp := t.tblFctApp)

(* mergeLidList: lid list -> lid list -> lid list *)
fun mergeLidList [] [] = []
  | mergeLidList [] l = l
  | mergeLidList l [] = l
  | mergeLidList (h1::t1) l =
    if L.mem h1 l then mergeLidList t1 l else h1::(mergeLidList t1 l)

(* mergeAb: 
   (lid list * ty) list -> (lid list * ty) list -> (lid list * ty) list *)
fun mergeAb [] [] = []
  | mergeAb [] l = l
  | mergeAb l [] = l
  | mergeAb ((h1,ty1)::t1) ((h2,ty2)::t2) =
    if ty1 = ty2 then (mergeLidList h1 h2, ty1)::(mergeAb t1 t2)
    else raise Fail "mergeAb"

(* mergeApAb: (lid list) list -> (lid list) list -> (lid list) list *)
fun mergeApAb [] [] = []
  | mergeApAb [] l = l
  | mergeApAb l [] = l
  | mergeApAb (h1::t1) (h2::t2) = (mergeLidList h1 h2)::(mergeApAb t1 t2)

(* addTbls: tables -> unit *)
fun addTbls t =
  (IM.iter (fn f (l,n,c,a,t) => fids := IM.add f (l,n,c,a,t) (!fids)) t.fids;
   IM.iter (fn l fs => L.iter (addF l) (IS.elements fs)) t.tblF;
   TM.iter (fn t fs => L.iter (addT t) (IS.elements fs)) t.tblT;
   IM.iter (fn f ls => addLa f (IS.elements ls)) t.tblLa;
   IM.iter (fn l f => tblLb := IM.add l f (!tblLb)) t.tblLb;
   IM.iter (fn l ls => L.iter (addE0 l) (IS.elements ls)) t.tblE;
   IM.iter (fn f ls =>
            tblAb := ((case IM.find f (!tblAb)
                         of args => IM.add f (mergeAb args ls) (!tblAb))
                      handle Not_found => IM.add f ls (!tblAb))) t.tblAb;
   IM.iter (fn l ls =>
            tblApAb := ((case IM.find l (!tblApAb)
                           of args => IM.add l (mergeApAb args ls) (!tblApAb))
                       handle Not_found => IM.add l ls (!tblApAb))) t.tblApAb;
   IM.iter (fn l l' => tblApL := IM.add l l' (!tblApL)) t.tblApL;
   IM.iter (fn l le => tblLid2le := IM.add l le (!tblLid2le)) t.tblLid2le;
   LM.iter (fn lc l => tblLoc2lid := LM.add lc l (!tblLoc2lid)) t.tblLoc2lid;
   LM.iter (fn lc s =>
            tblFctApp := ((case LM.find lc (!tblFctApp)
                             of args => raise Fail "FctApp in structure")
                          handle Not_found => LM.add lc s (!tblFctApp)))
           t.tblFctApp)

(* table1 - table2 *)
(* minusTbls: tables -> tables -> tables *)
fun minusTbls t1 t2 =
  let fun minusIMIS tbl1 tbl2 =
        IM.fold (fn id item m =>
                    ((case IM.find id tbl2
                        of item' => let val diff = IS.diff item item'
                                    in  if IS.is_empty diff then m
                                        else IM.add id diff m
                                    end)
                     handle Not_found => IM.add id item m)) tbl1 IM.empty
      fun minusIMO tbl1 tbl2 =
        IM.fold (fn id item m =>
                    ((case IM.find id tbl2 of _ => m)
                     handle Not_found => IM.add id item m)) tbl1 IM.empty
  in {fids = minusIMO t1.fids t2.fids,
      tblF = minusIMIS t1.tblF t2.tblF,
      tblT = TM.fold (fn t fs m =>
                      ((case TM.find t t2.tblT
                          of fs' => let val diff = IS.diff fs fs'
                                    in  if IS.is_empty diff then m
                                        else TM.add t diff m
                                    end)
                       handle Not_found => TM.add t fs m)) t1.tblT TM.empty,
      tblLa = minusIMIS t1.tblLa t2.tblLa,
      tblLb = minusIMO t1.tblLb t2.tblLb,
      tblE = minusIMIS t1.tblE t2.tblE,
      tblAb =
      IM.fold (fn id item m =>
               ((case IM.find id t2.tblAb
                   of item' =>
                      let val diff =
                              L.map2 (fn (l1,t) (l2,_) =>
                                      (L.filter (fn i => not(L.mem i l2)) l1,
                                       t)) item item'
                          fun empty [] = true
                            | empty lst =
                              L.for_all (fn ([],_) => true | _ => false) lst
                      in  if empty diff then m else IM.add id diff m
                      end)
                handle Not_found => IM.add id item m)) t1.tblAb IM.empty
      handle Invalid_argument _ => raise Fail "minusTbls",
      tblApAb =
      IM.fold (fn id item m =>
               ((case IM.find id t2.tblApAb
                   of item' =>
                      let val diff =
                              L.map2 (fn l1 l2 =>
                                      L.filter (fn i => not(L.mem i l2)) l1)
                                     item item'
                          fun empty [] = true
                            | empty lst =
                              L.for_all (fn [] => true | _ => false) lst
                      in  if empty diff then m else IM.add id diff m
                      end)
                handle Not_found => IM.add id item m)) t1.tblApAb IM.empty
      handle Invalid_argument _ => raise Fail "minusTbls",
      tblApL = IM.fold (fn l l' m =>
                           ((case IM.find l t2.tblApL of _ => m)
                            handle Not_found => IM.add l l' m))
                       t1.tblApL IM.empty,
      tblLid2le = minusIMO t1.tblLid2le t2.tblLid2le,
(*
      tblLoc2lid = LM.fold (fn c l m =>
                               ((case LM.find c t2.tblLoc2lid of _ => m)
                               handle Not_found => LM.add c l m))
                           t1.tblLoc2lid LM.empty,
*)
      tblLoc2lid = t1.tblLoc2lid,
      tblFctApp = LM.fold (fn c s m =>
                              ((case LM.find c t2.tblFctApp of _ => m)
                              handle Not_found => LM.add c s m))
                          t1.tblFctApp LM.empty}
  end

(* (f,f'), (l,l') \in substitution *)
(* substitute {f,l} \in "tables" to {f',l'} *)
(* if {f'',l''} \not\in substitution, *)
(* it is not a function type and thus deleted *)
(* substTbls: tables -> (fid * fid) list -> (lid * lid) list -> tables *)
fun substTbls t subf subl =
  let val fm = foldl (fn m (f, f') => IM.add f f' m) IM.empty subf
      val lm = foldl (fn m (l, l') => IM.add l l' m) IM.empty subl
      fun sf f = (IM.find f fm) handle Not_found => f
      fun sl l = (IM.find l lm) handle Not_found => l
      fun sle le = case le of Ve (vlid,f,i) => Ve (vlid, sf f, i)
                            | Ae l => Ae (sl l)
                            | FAe l => FAe (sl l)
                            | _ => le
      fun sfIS fs = IS.fold (fn f s => IS.add (sf f) s) fs IS.empty
      fun slIS ls = IS.fold (fn l s => IS.add (sl l) s) ls IS.empty
      fun sfFF s = L.map (fn (f,f') => (sf f, sf f')) s
      fun getPrev id map = IM.find id map handle Not_found => IS.empty
  in {fids = IM.fold (fn f (l,n,c,a,t) m =>
                         IM.add (sf f) (sl l,n,c,a,t) m) t.fids IM.empty,
      tblF = IM.fold (fn l fs m =>
                      let val newl = sl l
                      in  IM.add newl (IS.union (sfIS fs) (getPrev newl m)) m
                      end) t.tblF IM.empty,
      tblT = TM.fold (fn t fs m =>
                      let val prev = TM.find t m handle Not_found => IS.empty
                      in  TM.add t (IS.union (sfIS fs) prev) m
                      end) t.tblT TM.empty,
      tblLa = IM.fold (fn f ls m =>
                       let val newf = sf f
                       in  IM.add newf (IS.union (slIS ls) (getPrev newf m)) m
                       end) t.tblLa IM.empty,
      tblLb = IM.fold (fn l f m => IM.add (sl l) (sf f) m) t.tblLb IM.empty,
      tblE = IM.fold (fn l ls m =>
                      let val newl = sl l
                      in  IM.add newl (IS.union (slIS ls) (getPrev newl m)) m
                      end) t.tblE IM.empty,
      tblAb = IM.fold (fn f args m =>
                       let val newf = sf f
                           val newargs =
                               L.map (fn (ls,t) => (L.map sl ls, t)) args
                       in  (case IM.find newf m
                              of this => IM.add newf (mergeAb this newargs) m)
                           handle Not_found => IM.add newf newargs m
                       end) t.tblAb IM.empty,
      tblApAb = IM.fold (fn l args m =>
                         let val newl = sl l
                             val newargs = L.map (fn ls => L.map sl ls) args
                         in  (case IM.find newl m
                                of this =>
                                   IM.add newl (mergeApAb this newargs) m)
                             handle Not_found => IM.add newl newargs m
                         end) t.tblApAb IM.empty,
      tblApL = IM.fold (fn l l' m => IM.add (sl l) (sl l') m) t.tblApL
                       IM.empty,
      tblLid2le = IM.fold (fn l le m => IM.add (sl l) (sle le) m) t.tblLid2le
                          IM.empty,
      tblLoc2lid = LM.fold (fn loc l m => LM.add loc (sl l) m) t.tblLoc2lid
                           LM.empty,
      tblFctApp = LM.fold (fn loc s m =>
                           ((case LM.find loc m
                               of _ => raise Fail "FctApp in structure")
                            handle Not_found =>
                                   LM.add loc (sfFF s) m)) t.tblFctApp
                          LM.empty}
  end

(* prTbls: tables -> unit *)
fun prTbls t =
  (prFids (ref (t.fids)); prTblIS (ref (t.tblF)) "tblF"; 
   prTblIS (ref (t.tblLa)) "tblLa"; prTblI (ref (t.tblLb)) "tblLb";
   prTblIS (ref (t.tblE)) "tblE"; prTblAb (ref (t.tblAb));
   prTblIL (ref (t.tblApAb)) "tblApAb"; prTblI (ref (t.tblApL)) "tblApL";
   printLexp (ref (t.tblLid2le)); prLoc2I (ref (t.tblLoc2lid)) "tblLoc2lid")

(* initCfa: unit -> unit *)
fun initCfa () =
  (AU.init(); count := 0; flag := true; idl := 0; idf := 0; initftns := nil;
   fids := IM.empty; tblF := IM.empty; tblT := TM.empty;
   tblLa := IM.empty; tblLb := IM.empty; tblE := IM.empty;
   tblAb := IM.empty; tblApAb := IM.empty; tblApL := IM.empty;
   tblLid2le := IM.empty; tblLoc2lid := LM.empty; tblFctApp := LM.empty)

(* initCfaConstruct: basis * tables -> unit *)
fun initCfaConstruct (b as {c=C(_,ci),...}, t) =
  (basis0 := b; idl := ci.ll; idf := ci.lf; initftns := ci.fns; 
   fids := t.fids; tblF := t.tblF; tblT := t.tblT;
   tblLa := t.tblLa; tblLb := t.tblLb; tblE := t.tblE;
   tblAb := t.tblAb; tblApAb := t.tblApAb; tblApL := t.tblApL;
   tblLid2le := t.tblLid2le; tblLoc2lid := t.tblLoc2lid;
   tblFctApp := t.tblFctApp)

(* initialize t2fT *)
(* initCfaSolve: unit -> unit *)
fun initCfaSolve () =
  let val realfids = IM.fold (fn f _ s => IS.add f s) (!fids) IS.empty
  in  t2fT := TM.add (Ty.Var(Ty.Bound 0)) realfids TM.empty
  end

(* cfasE: unit -> unit *)
fun cfasE () =
  IM.iter (fn lid lids' =>
              IS.iter (fn lid' => IS.iter (fn f => addF' lid f)
                                          (getF0 lid')) lids') (!tblE)

(* appArg: (lid list * ty) list -> lid list list -> unit *)
fun appArg [] [] = ()
  | appArg ((lis',_)::resta) (lis::restb) =
    (L.iter (fn li =>
                if li > 0 then L.iter (fn li' => if li' > 0 then addE' li' li)
                                      lis') lis;
     appArg resta restb)
  | appArg [] _ = ()
  | appArg formal [] =
    L.iter (fn (lids,ty) =>
               L.iter (fn lid => if IS.is_empty (getF0 lid)
                                 then setF lid (ty2fids ty)) lids) formal

(* cfasAp: unit -> unit *)
fun cfasAp () =
  (IM.iter (fn lid1 actual =>
               IM.iter (fn fid formal =>
                           if IS.mem fid (getF0 lid1)
                           then appArg formal actual) (!tblAb)) (!tblApAb);
   IM.iter (fn lid lid1 =>
               IM.iter (fn fid bodies =>
                           if IS.mem fid (getF0 lid1)
                           then IS.iter (addE' lid) bodies) (!tblLa))
           (!tblApL))

(* cfasLb: unit -> unit *)
fun cfasLb () =
  IM.iter (fn lid fid => IS.iter (fn f' => addLa' fid (getLa0 f'))
                                 (getF0 lid)) (!tblLb)

end (* CfaUtils *)
