signature PERV =
  sig
    type fid = int
    val isPervasives: Absyn.exp -> bool
    val getPrimitives: string -> fid option
    val getPervasives: string -> fid option
    val getAri: fid -> fid
  end (* PERV *)

structure Perv : PERV =
  struct
    open Utils
    open AbsynUtils
    open Ty
    type fid = int

    (* getPrimitives: string -> fid option *)
    fun getPrimitives "+" = Some (-1)
      | getPrimitives "-" = Some (-3)
      | getPrimitives "*" = Some (-5)
      | getPrimitives "/" = Some (-7)
      | getPrimitives "unary_minus" = Some (-9)
      | getPrimitives "%" = Some (-10)
      | getPrimitives "**" = Some (-12)
      | getPrimitives "++" = Some (-14)
      | getPrimitives "--" = Some (-15)
      | getPrimitives "+=" = Some (-16)
      | getPrimitives "-=" = Some (-18)
      | getPrimitives "*=" = Some (-20)
      | getPrimitives "/=" = Some (-22)
      | getPrimitives "<<" = Some (-24)
      | getPrimitives ">>" = Some (-26)
      | getPrimitives "not" = Some (-28)
      | getPrimitives "=" = Some (-29)
      | getPrimitives "<>" = Some (-31)
      | getPrimitives "<" = Some (-33)
      | getPrimitives "<=" = Some (-35)
      | getPrimitives ">" = Some (-37)
      | getPrimitives ">=" = Some (-39)
      | getPrimitives "@" = Some (-41)
      | getPrimitives "^" = Some (-43)
      | getPrimitives "::" = Some (-45)
      | getPrimitives "andalso" = Some (-46)
      | getPrimitives "orelse" = Some (-48)
      | getPrimitives _ = None

    val pervMap = SH.create 127
    val _ = List.iter (fn (i,n) => SH.add pervMap n i)
        [
(*
         (* -1 ~ -49 : primitive operators *)
         (-1,"+"),(-3,"-"),(-5,"*"),(-7,"/"),(-9,"unary_minus"),
         (-10,"%"),(-12,"**"),(-14,"++"),(-15,"--"),
         (-16,"+="),(-18,"-="),(-20,"*="),(-22,"/="),
         (-24,"<<"),(-26,">>"),(-28,"not"),(-29,"="),(-31,"<>"),
         (-33,"<"),(-35,"<="),(-37,">"),(-39,">="),
         (-41,"@"),(-43,"^"),(-45,"::"),
         (-46,"andalso"),(-48,"orelse"),
(*
         (-32,"andalso"),(-33,"&&"),(-34,"orelse"),(-35,"||"),
*)
*)
         (* -51 ~ -181 : Pervasives functions *)
         (-51,"Pervasives.invalid_arg"),(-52,"Pervasives.failwith"),
         (-53,"Pervasives.compare"),(-55,"Pervasives.min"),
         (-57,"Pervasives.max"),(-59,"Pervasives.or"),
         (-61,"Pervasives.succ"),(-62,"Pervasives.pred"),
         (-63,"Pervasives.abs"),(-64,"Pervasives.lnot"),
         (-65,"Pervasives.asr"),(-67,"Pervasives.sqrt"),
         (-68,"Pervasives.exp"),(-69,"Pervasives.log"),
         (-70,"Pervasives.log10"),(-71,"Pervasives.cos"),
         (-72,"Pervasives.sin"),(-73,"Pervasives.tan"),
         (-74,"Pervasives.acos"),(-75,"Pervasives.asin"),
         (-76,"Pervasives.atan"),(-77,"Pervasives.atan2"),
         (-79,"Pervasives.cosh"),(-80,"Pervasives.sinh"),
         (-81,"Pervasives.tanh"),(-82,"Pervasives.ceil"),
         (-83,"Pervasives.floor"),(-84,"Pervasives.abs_float"),
         (-85,"Pervasives.mod_float"),(-87,"Pervasives.frexp"),
         (-88,"Pervasives.ldexp"),(-90,"Pervasives.modf"),
         (-91,"Pervasives.float"),(-92,"Pervasives.float_of_int"),
         (-93,"Pervasives.truncate"),(-94,"Pervasives.int_of_float"),
         (-95,"Pervasives.int_of_char"),(-96,"Pervasives.char_of_int"),
         (-97,"Pervasives.ignore"),(-98,"Pervasives.string_of_bool"),
         (-99,"Pervasives.bool_of_string"),(-100,"Pervasives.string_of_int"),
         (-101,"Pervasives.int_of_string"),
         (-102,"Pervasives.string_of_float"),
         (-103,"Pervasives.float_of_string"),(-104,"Pervasives.fst"),
         (-105,"Pervasives.snd"),(-106,"Pervasives.print_char"),
         (-107,"Pervasives.print_string"),(-108,"Pervasives.print_int"),
         (-109,"Pervasives.print_float"),(-110,"Pervasives.print_endline"),
         (-111,"Pervasives.print_newline"),(-112,"Pervasives.prerr_char"),
         (-113,"Pervasives.prerr_string"),(-114,"Pervasives.prerr_int"),
         (-115,"Pervasives.prerr_float"),(-116,"Pervasives.prerr_endline"),
         (-117,"Pervasives.prerr_newline"),(-118,"Pervasives.read_line"),
         (-119,"Pervasives.read_int"),(-120,"Pervasives.read_float"),
         (-121,"Pervasives.open_out"),(-122,"Pervasives.open_out_bin"),
         (-123,"Pervasives.open_out_gen"),(-126,"Pervasives.flush"),
         (-127,"Pervasives.output_char"),(-129,"Pervasives.output_string"),
         (-131,"Pervasives.output"),(-135,"Pervasives.output_byte"),
         (-137,"Pervasives.output_binary_int"),
         (-139,"Pervasives.output_value"),
         (-141,"Pervasives.seek_out"),(-143,"Pervasives.pos_out"),
         (-144,"Pervasives.out_channel_length"),(-145,"Pervasives.close_out"),
         (-146,"Pervasives.set_binary_mode_out"),(-148,"Pervasives.open_in"),
         (-149,"Pervasives.open_in_bin"),(-150,"Pervasives.open_in_gen"),
         (-153,"Pervasives.input_char"),(-154,"Pervasives.input_line"),
         (-155,"Pervasives.input"),(-159,"Pervasives.really_input"),
         (-163,"Pervasives.input_byte"),(-164,"Pervasives.input_binary_int"),
         (-165,"Pervasives.input_value"),(-166,"Pervasives.seek_in"),
         (-168,"Pervasives.pos_in"),(-169,"Pervasives.in_channel_length"),
         (-170,"Pervasives.close_in"),(-171,"Pervasives.set_binary_mode_in"),
         (-173,"Pervasives.incr"),(-174,"Pervasives.decr"),
         (-175,"Pervasives.exit"),(-176,"Pervasives.at_exit"),
         (-177,"Pervasives.unsafe_really_input"),
         (-181,"Pervasives.do_at_exit"),
         (* -201 ~ -255 : Array functions *)
         (-201,"Array.length"),(-202,"Array.get"),(-204,"Array.set"),
         (-207,"Array.make"),(-209,"Array.create"),(-211,"Array.init"),
         (-213,"Array.make_matrix"),(-216,"Array.create_matrix"),
         (-219,"Array.append"),(-221,"Array.concat"),(-222,"Array.sub"),
         (-225,"Array.copy"),(-226,"Array.fill"),(-230,"Array.blit"),
         (-235,"Array.to_list"),(-236,"Array.of_list"),(-237,"Array.iter"),
         (-239,"Array.map"),(-241,"Array.iteri"),(-243,"Array.mapi"),
         (-245,"Array.fold_left"),(-248,"Array.fold_right"),
         (-251,"Array.unsafe_get"),(-253,"Array.unsafe_set"),
         (* -261 ~ -338 : List functions *)
         (-261,"List.length"),(-262,"List.hd"),(-263,"List.tl"),
         (-264,"List.nth"),(-266,"List.rev"),(-267,"List.append"),
         (-269,"List.rev_append"),(-271,"List.concat"),(-272,"List.flatten"),
         (-273,"List.iter"),(-275,"List.map"),(-277,"List.rev_map"),
         (-279,"List.fold_left"),(-282,"List.fold_right"),(-285,"List.iter2"),
         (-288,"List.map2"),(-291,"List.rev_map2"),(-294,"List.fold_left2"),
         (-298,"List.fold_right2"),(-302,"List.for_all"),(-304,"List.exists"),
         (-306,"List.for_all2"),(-309,"List.exists2"),(-312,"List.mem"),
         (-314,"List.memq"),(-316,"List.find"),(-318,"List.filter"),
         (-320,"List.find_all"),(-322,"List.partition"),(-324,"List.assoc"),
         (-326,"List.assq"),(-328,"List.mem_assoc"),(-330,"List.mem_assq"),
         (-332,"List.remove_assoc"),(-334,"List.remove_assq"),
         (-336,"List.split"),(-337,"List.combine"),
         (* -341 ~ -401 : String functions *)
         (-341,"String.length"),(-342,"String.get"),(-344,"String.set"),
         (-347,"String.create"),(-348,"String.make"),(-350,"String.copy"),
         (-351,"String.sub"),(-354,"String.fill"),(-358,"String.blit"),
         (-363,"String.concat"),(-365,"String.escaped"),(-366,"String.index"),
         (-368,"String.rindex"),(-370,"String.index_from"),
         (-373,"String.rindex_from"),(-376,"String.contains"),
         (-378,"String.contains_from"),(-381,"String.rcontains_from"),
         (-384,"String.uppercase"),(-385,"String.lowercase"),
         (-386,"String.capitalize"),(-387,"String.uncapitalize"),
         (-388,"String.unsafe_get"),(-390,"String.unsafe_set"),
         (-393,"String.unsafe_blit"),(-398,"String.unsafe_fill"),
         (* -411 ~ -416 : Char functions *)
         (-411,"Char.code"),(-412,"Char.chr"),(-413,"Char.escaped"),
         (-414,"Char.lowercase"),(-415,"Char.uppercase"),
         (-416,"Char.unsafe_chr"),
         (* -421 ~ -425 : Arg functions *)
         (-421,"Arg.parse"),(-424,"Arg.usage"),
         (* -431 ~ -450 : Buffer functions *)
         (-431,"Buffer.create"),(-432,"Buffer.contents"),
         (-433,"Buffer.length"),(-434,"Buffer.clear"),
         (-435,"Buffer.reset"),(-436,"Buffer.add_char"),
         (-438,"Buffer.add_string"),
         (-440,"Buffer.add_substring"),
         (-444,"Buffer.add_buffer"),
         (-446,"Buffer.add_channel"),
         (-449,"Buffer.output_buffer"),
         (* -451 ~ -454 : Callback functions *)
         (-451,"Callback.register"),(-453,"Callback.register_exception"),
         (* -461 ~ -470 : Digest functions *)
         (-461,"Digest.string"),(-462,"Digest.substring"),
         (-465,"Digest.channel"),(-467,"Digest.file"),(-468,"Digest.output"),
         (-470,"Digest.input"),
         (* -471 ~ -483 : Filename functions *)
         (-471,"Filename.concat"),(-473,"Filename.is_relative"),
         (-474,"Filename.is_implicit"),(-475,"Filename.check_suffix"),
         (-477,"Filename.chop_suffix"),(-479,"Filename.chop_extension"),
         (-480,"Filename.basename"),(-481,"Filename.dirname"),
         (-482,"Filename.temp_file"),
         (* -491 ~ -498 : Gc functions *)
         (-491,"Gc.stat"),(-492,"Gc.print_stat"),(-493,"Gc.get"),
         (-494,"Gc.set"),(-495,"Gc.minor"),(-496,"Gc.major"),
         (-497,"Gc.full_major"),(-498,"Gc.compact"),
         (* -499 : Lazy function *)
         (-499,"Lazy.force"),
         (* -500 ~ -510 : Lexing functions *)
         (-500,"Lexing.from_channel"),(-501,"Lexing.from_string"),
         (-502,"Lexing.from_function"),(-503,"Lexing.lexeme"),
         (-504,"Lexing.lexeme_char"),(-506,"Lexing.lexeme_start"),
         (-507,"Lexing.lexeme_end"),(-508,"Lexing.engine"),
         (* -511 ~ -527 : Marshal functions *)
         (-511,"Marshal.to_channel"),(-514,"Marshal.to_string"),
         (-516,"Marshal.to_buffer"),(-521,"Marshal.from_channel"),
         (-522,"Marshal.from_string"),(-524,"Marshal.data_size"),
         (-526,"Marshal.total_size"),
         (* -531 ~ -549 : Obj functions *)
         (-531,"Obj.repr"),(-532,"Obj.obj"),(-533,"Obj.magic"),
         (-534,"Obj.is_block"),(-535,"Obj.tag"),(-536,"Obj.size"),
         (-537,"Obj.field"),(-539,"Obj.set_field"),(-542,"Obj.new_block"),
         (-544,"Obj.dup"),(-545,"Obj.truncate"),(-547,"Obj.marshal"),
         (-548,"Obj.unmarshal"),
         (* -551 ~ 586 : Oo functions *)
         (-551,"Oo.copy"),(-552,"Oo.new_method"),(-553,"Oo.get_class"),
         (-555,"Oo.new_variable"),(-557,"Oo.get_variable"),
         (-559,"Oo.copy_variables"),(-563,"Oo.get_method_label"),
         (-565,"Oo.get_method"),(-567,"Oo.set_method"),(-570,"Oo.narrow"),
         (-574,"Oo.widen"),(-575,"Oo.add_initializer"),
         (-577,"Oo.create_table"),(-578,"Oo.init_class"),
         (-579,"Oo.create_object"),(-580,"Oo.run_initializers"),
         (-582,"Oo.object_from_struct"),(-583,"Oo.send"),
         (-585,"Oo.stats"),(-586,"Oo.show_buckets"),
         (* -591 ~ -595 : Printexc functions *)
         (-591,"Printexc.catch"),(-593,"Printexc.print"),
         (-595,"Printexc.to_string"),
         (* -601 ~ -609 : Queue functions *)
         (-601,"Queue.create"),(-602,"Queue.add"),(-604,"Queue.take"),
         (-605,"Queue.peek"),(-606,"Queue.clear"),(-607,"Queue.length"),
         (-608,"Queue.iter"),
         (* -611 ~ -616 : Random functions *)
         (-611,"Random.init"),(-612,"Random.full_init"),
         (-613,"Random.self_init"),(-614,"Random.bits"),
         (-615,"Random.int"),(-616,"Random.float"),
         (* -621 ~ -627 : Sort functions *)
         (-621,"Sort.list"),(-623,"Sort.array"),(-625,"Sort.merge"),
         (* -631 ~ -638 : Stack functions *)
         (-631,"Stack.create"),(-632,"Stack.push"),(-634,"Stack.pop"),
         (-635,"Stack.clear"),(-636,"Stack.length"),(-637,"Stack.iter"),
         (* -641 ~ -666 : Stream functions *)
         (-641,"Stream.from"),(-642,"Stream.of_list"),
         (-643,"Stream.of_string"),(-644,"Stream.of_channel"),
         (-645,"Stream.iter"),(-647,"Stream.next"),(-648,"Stream.empty"),
         (-649,"Stream.peek"),(-650,"Stream.junk"),(-651,"Stream.count"),
         (-652,"Stream.npeek"),(-654,"Stream.iapp"),(-656,"Stream.icons"),
         (-658,"Stream.ising"),(-659,"Stream.lapp"),(-661,"Stream.lcons"),
         (-663,"Stream.lsing"),(-664,"Stream.slazy"),(-665,"Stream.dump"),
         (* -671 ~ -684 : Sys functions *)
         (-671,"Sys.file_exists"),(-672,"Sys.remove"),(-673,"Sys.rename"),
         (-675,"Sys.getenv"),(-676,"Sys.command"),(-677,"Sys.time"),
         (-678,"Sys.chdir"),(-679,"Sys.getcwd"),(-680,"Sys.signal"),
         (-682,"Sys.set_signal"),(-684,"Sys.catch_break"),
         (* -691 ~ -708 : Weak functions *)
         (-691,"Weak.create"),(-692,"Weak.length"),(-693,"Weak.set"),
         (-696,"Weak.get"),(-698,"Weak.check"),(-700,"Weak.fill"),
         (-704,"Weak.blit"),
         (* -711 ~ -717 : Printf functions *)
         (-711,"Printf.fprintf"),(-713,"Printf.printf"),
         (-714,"Printf.deprintf"),(-715,"Printf.sprintf"),
         (-716,"Printf.bprintf"),
         (* -718 : Genlex functions *)
         (-718,"Genlex.make_lexer"),
         (* -721 ~ -733 : Parsing functions *)
         (-721,"Parsing.symbol_start"),(-722,"Parsing.symbol_end"),
         (-723,"Parsing.rhs_start"),(-724,"Parsing.rhs_end"),
         (-725,"Parsing.clear_parser"),(-726,"Parsing.yyparse"),
         (-730,"Parsing.peek_val"),(-732,"Parsing.is_current_lookahead"),
         (-733,"Parsing.parse_error"),
         (* -741 ~ -884 : Format functions *)
         (-741,"Format.open_box"),(-742,"Format.close_box"),
         (-743,"Format.print_string"),(-744,"Format.print_as"),
         (-746,"Format.print_int"),(-747,"Format.print_float"),
         (-748,"Format.print_char"),(-749,"Format.print_bool"),
         (-750,"Format.print_space"),(-751,"Format.print_cut"),
         (-752,"Format.print_break"),(-754,"Format.print_flush"),
         (-755,"Format.print_newline"),(-756,"Format.force_newline"),
         (-757,"Format.print_if_newline"),(-758,"Format.set_margin"),
         (-759,"Format.get_margin"),(-760,"Format.set_max_indent"),
         (-761,"Format.get_max_indent"),(-762,"Format.set_max_boxes"),
         (-763,"Format.get_max_boxes"),(-764,"Format.over_max_boxes"),
         (-765,"Format.open_hbox"),(-766,"Format.open_vbox"),
         (-767,"Format.open_hvbox"),(-768,"Format.open_hovbox"),
         (-769,"Format.open_tbox"),(-770,"Format.close_tbox"),
         (-771,"Format.print_tbreak"),(-773,"Format.set_tab"),
         (-774,"Format.print_tab"),(-775,"Format.set_ellipsis_text"),
         (-776,"Format.get_ellipsis_text"),
         (-777,"Format.set_formatter_out_channel"),
         (-778,"Format.set_formatter_output_functions"),
         (-780,"Format.get_formatter_output_functions"),
         (-781,"Format.set_all_formatter_output_functions"),
         (-785,"Format.get_all_formatter_output_functions"),
         (-786,"Format.formatter_of_out_channel"),
         (-787,"Format.formatter_of_buffer"),
         (-788,"Format.flush_str_formatter"),
         (-789,"Format.make_formatter"),(-791,"Format.pp_open_hbox"),
         (-793,"Format.pp_open_vbox"),(-795,"Format.pp_open_hvbox"),
         (-797,"Format.pp_open_hovbox"),(-799,"Format.pp_open_box"),
         (-801,"Format.pp_close_box"),(-803,"Format.pp_print_string"),
         (-805,"Format.pp_print_as"),(-808,"Format.pp_print_int"),
         (-810,"Format.pp_print_float"),(-812,"Format.pp_print_char"),
         (-814,"Format.pp_print_bool"),(-816,"Format.pp_print_break"),
         (-819,"Format.pp_print_cut"),(-821,"Format.pp_print_space"),
         (-823,"Format.pp_force_newline"),(-825,"Format.pp_print_flush"),
         (-827,"Format.pp_print_newline"),(-829,"Format.pp_print_if_newline"),
         (-831,"Format.pp_open_tbox"),(-833,"Format.pp_close_tbox"),
         (-835,"Format.pp_print_tbreak"),(-838,"Format.pp_set_tab"),
         (-840,"Format.pp_print_tab"),(-842,"Format.pp_set_margin"),
         (-844,"Format.pp_get_margin"),(-846,"Format.pp_set_max_indent"),
         (-848,"Format.pp_get_max_indent"),(-850,"Format.pp_set_max_boxes"),
         (-852,"Format.pp_get_max_boxes"),(-854,"Format.pp_over_max_boxes"),
         (-856,"Format.pp_set_ellipsis_text"),
         (-858,"Format.pp_get_ellipsis_text"),
         (-860,"Format.pp_set_formatter_out_channel"),
         (-862,"Format.pp_set_formatter_output_functions"),
         (-865,"Format.pp_get_formatter_output_functions"),
         (-867,"Format.pp_set_all_formatter_output_functions"),
         (-872,"Format.pp_get_all_formatter_output_functions"),
         (-874,"Format.fprintf"),(-876,"Format.bprintf"),
         (-878,"Format.printf"),(-879,"Format.eprintf"),
         (-880,"Format.sprintf"),
         (* -881 ~ -1102 : Unix functions *)
         (-881,"Unix.error_message"),(-882,"Unix.handle_unix_error"),
         (-884,"Unix.environment"),(-885,"Unix.getenv"),
         (-886,"Unix.putenv"),(-888,"Unix.execv"),(-890,"Unix.execve"),
         (-893,"Unix.execvp"),(-895,"Unix.execvpe"),(-898,"Unix.fork"),
         (-899,"Unix.wait"),(-900,"Unix.waitpid"),(-902,"Unix.system"),
         (-903,"Unix.getpid"),(-904,"Unix.getppid"),(-905,"Unix.nice"),
         (-906,"Unix.openfile"),(-909,"Unix.close"),(-910,"Unix.read"),
         (-914,"Unix.write"),(-918,"Unix.in_channel_of_descr"),
         (-919,"Unix.out_channel_of_descr"),(-920,"Unix.descr_of_in_channel"),
         (-921,"Unix.descr_of_out_channel"),(-922,"Unix.lseek"),
         (-925,"Unix.truncate"),(-927,"Unix.ftruncate"),(-929,"Unix.stat"),
         (-930,"Unix.lstat"),(-931,"Unix.fstat"),(-932,"Unix.unlink"),
         (-933,"Unix.rename"),(-935,"Unix.link"),(-937,"Unix.chmod"),
         (-939,"Unix.fchmod"),(-941,"Unix.chown"),(-944,"Unix.fchown"),
         (-947,"Unix.umask"),(-948,"Unix.access"),(-950,"Unix.dup"),
         (-951,"Unix.dup2"),(-953,"Unix.set_nonblock"),
         (-954,"Unix.clear_nonblock"),(-955,"Unix.set_close_on_exec"),
         (-956,"Unix.clear_close_on_exec"),(-957,"Unix.mkdir"),
         (-959,"Unix.rmdir"),(-960,"Unix.chdir"),(-961,"Unix.getcwd"),
         (-962,"Unix.chroot"),(-963,"Unix.opendir"),(-964,"Unix.readdir"),
         (-965,"Unix.rewinddir"),(-966,"Unix.closedir"),(-967,"Unix.pipe"),
         (-968,"Unix.mkfifo"),(-970,"Unix.create_process"),
         (-975,"Unix.create_process_env"),(-981,"Unix.open_process_in"),
         (-982,"Unix.open_process_out"),(-983,"Unix.open_process:"),
         (-984,"Unix.open_process_full"),(-986,"Unix.close_process_in"),
         (-987,"Unix.close_process_out"),(-988,"Unix.close_process"),
         (-989,"Unix.close_process_full"),(-990,"Unix.symlink"),
         (-992,"Unix.readlink"),(-993,"Unix.select"),(-997,"Unix.lockf"),
         (-1000,"Unix.kill"),(-1002,"Unix.sigprocmask"),
         (-1004,"Unix.sigpending"),(-1005,"Unix.sigsuspend"),
         (-1006,"Unix.pause"),(-1007,"Unix.time"),(-1008,"Unix.gettimeofday"),
         (-1009,"Unix.gmtime"),(-1010,"Unix.localtime"),(-1011,"Unix.mktime"),
         (-1012,"Unix.alarm"),(-1013,"Unix.sleep"),(-1014,"Unix.times"),
         (-1015,"Unix.utimes"),(-1018,"Unix.getitimer"),
         (-1019,"Unix.setitimer"),(-1021,"Unix.getuid"),
         (-1022,"Unix.geteuid"),(-1023,"Unix.setuid"),(-1024,"Unix.getgid"),
         (-1025,"Unix.getegid"),(-1026,"Unix.setgid"),
         (-1027,"Unix.getgroups"),(-1028,"Unix.getlogin"),
         (-1029,"Unix.getpwnam"),(-1030,"Unix.getgrnam"),
         (-1031,"Unix.getpwuid"),(-1032,"Unix.getgrgid"),
         (-1033,"Unix.inet_addr_of_string"),
         (-1034,"Unix.string_of_inet_addr"),(-1035,"Unix.socket"),
         (-1038,"Unix.socketpair"),(-1041,"Unix.accept"),(-1042,"Unix.bind"),
         (-1044,"Unix.connect"),(-1046,"Unix.listen"),(-1048,"Unix.shutdown"),
         (-1050,"Unix.getsockname"),(-1051,"Unix.getpeername"),
         (-1052,"Unix.recv"),(-1057,"Unix.recvfrom"),(-1062,"Unix.send"),
         (-1067,"Unix.sendto"),(-1073,"Unix.getsockopt"),
         (-1075,"Unix.setsockopt"),(-1078,"Unix.open_connection"),
         (-1079,"Unix.shutdown_connection"),(-1080,"Unix.establish_server"),
         (-1082,"Unix.gethostname"),(-1083,"Unix.gethostbyname"),
         (-1084,"Unix.gethostbyaddr"),(-1085,"Unix.getprotobyname"),
         (-1086,"Unix.getprotobynumber"),(-1087,"Unix.getservbyname"),
         (-1089,"Unix.getservbyport"),(-1091,"Unix.tcgetattr"),
         (-1092,"Unix.tcsetattr"),(-1095,"Unix.tcsendbreak"),
         (-1097,"Unix.tcdrain"),(-1098,"Unix.tcflush"),(-1100,"Unix.tcflow"),
         (-1102,"Unix.setsid")] 

    (* isPrimitives: string -> bool *)
    fun isPrimitives name =
      List.mem name ["+", "-", "*", "/", "unary_minus", "%", "**", "++", "--",
                     "+=", "-=", "*=", "/=", "<<", ">>", "not", "=", "<>",
                     "<", "<=", ">", ">=", "@", "^", "::", "andalso","orelse"]

    (* isPervasives: Absyn.exp -> bool *)
    fun isPervasives exp =
      case exp
        of Absyn.VarExp (vlid,_,_) =>
           ((case SH.find pervMap (varLName vlid) of _ => true)
            handle Not_found => isPrimitives (varLName vlid))
         | Absyn.AppExp (exp1,_,_,_) => isPervasives exp1
         | _ => false

    (* getPervasives: string -> fid option *)
    fun getPervasives fname =
      (Some(SH.find pervMap fname)) handle Not_found => None

    val ari1 = List.fold_left (fn res f => IS.add f res) IS.empty
        [-9,-14,-15,-28,-55,
         -51,-52,-61,-62,-63,-64,-67,-68,-69,-70,-71,-72,-73,-74,-75,
         -76,-79,-80,-81,-82,-83,-84,-87,-126,-143,-144,-145,-148,-149,
         -153,-154,-163,-164,-165,-168,-169,-170,-173,-174,-175,-176,-181,
         -201,-221,-225,-235,-236,-261,-262,-263,-266,-271,-272,-336,
         -341,-347,-350,-365,-384,-385,-386,-387,-411,-412,-413,-414,-415,
         -416,-431,-432,-433,-434,-435,-461,-467,-470,-473,-474,-479,-480,
         -481,-506,-507,-521,-531,-532,-533,-534,-535,-536,-544,-547,-551,
         -552,-574,-577,-578,-579,-582,-585,-586,-595,-601,-604,-605,-606,
         -607,-611,-612,-613,-614,-615,-616,-631,-634,-635,-636,-641,-642,
         -643,-644,-647,-648,-649,-650,-651,-658,-663,-664,-671,-672,-675,
         -676,-677,-678,-679,-684,-691,-692,-713,-714,-715,-718,-721,-722,
         -723,-724,-725,-732,-733,-741,-742,-743,-746,-747,-748,-749,-750,
         -751,-754,-755,-756,-757,-758,-759,-760,-761,-762,-763,-764,-765,
         -766,-767,-768,-769,-770,-773,-774,-775,-776,-777,-780,-785,-786,
         -787,-788,-878,-879,-880,-881,-884,-885,-898,-899,-902,-903,-904,
         -905,-909,-918,-919,-920,-921,-929,-930,-931,-932,-947,-950,-953,
         -954,-955,-956,-959,-960,-961,-962,-963,-964,-965,-966,-967,-981,
         -982,-983,-986,-987,-988,-989,-992,-1004,-1005,-1006,-1007,-1008,
         -1009,-1010,-1011,-1012,-1013,-1014,-1018,-1021,-1022,-1023,-1024,
         -1025,-1026,-1027,-1028,-1029,-1030,-1031,-1032,-1033,-1034,-1041,
         -1050,-1051,-1078,-1079,-1082,-1083,-1084,-1085,-1086,-1091,-1097,
         -1102]
    val ari2 = List.fold_left (fn res f => IS.add f res) IS.empty
        [-1,-3,-5,-7,-10,-12,-16,-18,-20,-22,-24,-26,-29,-31,-33,-35,-37,
         -39,-41,-43,-46,-48,
         -53,-55,-57,-59,-65,-77,-85,-88,-127,-129,-135,
         -137,-139,-141,-146,-166,-171,
         -202,-207,-209,-211,-219,-237,-239,-241,-243,-251,-264,-267,-269,
         -273,-275,-277,-302,-304,-312,-314,-316,-318,-320,-322,-324,-326,
         -328,-330,-332,-334,-337,-342,-348,-363,-366,-368,-376,-388,-424,
         -436,-438,-444,-449,-451,-453,-465,-468,-471,-475,-477,-482,-504,
         -514,-522,-524,-526,-537,-542,-545,-548,-553,-555,-557,-563,-565,
         -575,-580,-583,-591,-593,-602,-608,-621,-623,-632,-637,-645,-652,
         -654,-656,-659,-661,-665,-673,-680,-682,-696,-698,-711,-716,-730,
         -744,-752,-771,-778,-789,-791,-793,-795,-797,-799,-801,-803,-808,
         -810,-812,-814,-819,-821,-823,-825,-827,-829,-831,-833,-838,-840,
         -842,-844,-846,-848,-850,-852,-854,-856,-858,-860,-865,-872,-874,
         -876,-882,-886,-888,-893,-900,-925,-927,-933,-935,-937,-939,-948,
         -951,-957,-968,-984,-990,-1000,-1002,-1019,-1042,-1044,-1046,-1048,
         -1073,-1080,-1087,-1089,-1095,-1098,-1100]
    val ari3 = List.fold_left (fn res f => IS.add f res) IS.empty
        [-123,-150,-204,-213,-216,-222,-245,-248,-253,-279,-282,-285,-288,
         -291,-306,-309,-344,-351,-370,-373,-378,-381,-390,-421,-446,-462,
         -508,-511,-539,-567,-625,-693,-805,-816,-835,-862,-890,-895,-906,
         -922,-941,-944,-997,-1015,-1035,-1038,-1075,-1092]
    val ari4 = List.fold_left (fn res f => IS.add f res) IS.empty
        [-131,-155,-159,-177,-226,-294,-298,-354,-398,-440,-559,-570,-700,
         -726,-781,-910,-914,-993]
    val ari5 = List.fold_left (fn res f => IS.add f res) IS.empty
        [-230,-358,-393,-516,-704,-867,-970,-1052,-1057,-1062]
    val ari6 = List.fold_left (fn res f => IS.add f res) IS.empty
        [-975,-1067]

    (* getAri: fid -> int *)
    fun getAri fid =
      if fid > -51 then 1
      else if fid <= -90 andalso fid >= -122 then 1
      else if fid <= -491 andalso fid >= -503 then 1
      else if IS.mem fid ari1 then 1
      else if IS.mem fid ari2 then 2
      else if IS.mem (fid+1) ari2 then 1
      else if IS.mem fid ari3 then 3
      else if IS.mem (fid+1) ari3 then 2
      else if IS.mem (fid+2) ari3 then 1
      else if IS.mem fid ari4 then 4
      else if IS.mem (fid+1) ari4 then 3
      else if IS.mem (fid+2) ari4 then 2
      else if IS.mem (fid+3) ari4 then 1
      else if IS.mem fid ari5 then 5
      else if IS.mem (fid+1) ari5 then 4
      else if IS.mem (fid+2) ari5 then 3
      else if IS.mem (fid+3) ari5 then 2
      else if IS.mem (fid+4) ari5 then 1
      else if IS.mem fid ari6 then 6
      else if IS.mem (fid+1) ari6 then 5
      else if IS.mem (fid+2) ari6 then 4
      else if IS.mem (fid+3) ari6 then 3
      else if IS.mem (fid+4) ari6 then 2
      else if IS.mem (fid+4) ari6 then 1
      else raise Fail "getAri"

  end (* Perv *)
